import os 


def check (cmd ,mf ):
    m =mf .findNode ("rtree")
    if m is None or m .filename is None :
        return None 

    try :
        rtree_dylibs =os .scandir (os .path .join (m .packagepath [0 ],"lib"))
    except OSError :
        return None 

    frameworks =[lib .path for lib in rtree_dylibs ]

    return {"frameworks":frameworks }



import sys 

from modulegraph .modulegraph import MissingModule 


def check (cmd ,mf ):
    if sys .version_info [0 ]!=2 :
        return {}



    m =mf .findNode ("_xmlplus")
    if m is not None and isinstance (m ,MissingModule ):
        mf .removeReference (mf .findNode ("xml"),m )

    return {}



import sys 
import os 
import numpy as np 
import json 
import csv 
from PySide6 .QtWidgets import (QApplication ,QMainWindow ,QVBoxLayout ,QHBoxLayout ,QWidget ,QPushButton ,QComboBox ,QListWidget ,QLabel ,QTableWidget ,QTableWidgetItem ,QFileDialog ,QScrollArea ,QSplitter ,QTabWidget ,QLineEdit ,QCheckBox ,QMessageBox ,QListWidgetItem ,QSpinBox ,QDialog ,QTextEdit ,QStyleFactory ,QGroupBox ,QGridLayout ,QFormLayout ,QDoubleSpinBox ,QLayout ,QLayoutItem )
from PySide6 .QtGui import QPalette ,QColor 
from PySide6 .QtCore import Qt 
from PySide6 .QtGui import QColor 
import matplotlib .pyplot as plt 
from matplotlib .backends .backend_qt5agg import FigureCanvasQTAgg as FigureCanvas 
import copy 
from matplotlib .backends .backend_qt5agg import NavigationToolbar2QT as NavigationToolbar 


class SDDatabaseViewer (QMainWindow ):
    def __init__ (self ):
        super ().__init__ ()
        self .setWindowTitle ("SD Database Viewer")
        self .setGeometry (100 ,100 ,1200 ,800 )

        self .central_widget =QWidget ()
        self .setCentralWidget (self .central_widget )
        self .layout =QHBoxLayout (self .central_widget )

        self .loaded_data =None 
        self .original_data =None 
        self .current_folder =None 
        self .comparison_data ={}
        self .is_updating_plot_options =False 
        self .ml_variables =[]
        self .plot_options ={}

        self .plot_fitted_data_checkbox =QCheckBox ("Plot Fitted Data")
        self .plot_baseline_checkbox =QCheckBox ("Plot Baseline")

        self .file_type =None 
        self .plot_options_form =None 
        self .setup_ui ()
        self .setup_plotting_tab ()

    def setup_ui (self ):

        left_panel =QWidget ()
        left_layout =QVBoxLayout (left_panel )

        self .app_name_label =QLabel ("SD Database Viewer")
        self .app_name_label .setAlignment (Qt .AlignmentFlag .AlignCenter )
        self .app_name_label .setStyleSheet ("font-size: 22px; font-weight: bold;")
        self .email_label =QLabel ("shankar.dutt@anu.edu.au")
        self .email_label .setAlignment (Qt .AlignmentFlag .AlignCenter )

        left_layout .addWidget (self .app_name_label )
        left_layout .addWidget (self .email_label )

        self .file_type_combo =QComboBox ()
        self .file_type_combo .addItems ([".event_fitting.npz",".dataset.npz",".MLdataset.npz",".event_data.npz"])
        self .file_type_combo .currentIndexChanged .connect (self .update_file_list )
        left_layout .addWidget (QLabel ("Select File Type:"))
        left_layout .addWidget (self .file_type_combo )

        self .folder_button =QPushButton ("Select Folder")
        self .folder_button .clicked .connect (self .select_folder )
        left_layout .addWidget (self .folder_button )

        self .file_list =QListWidget ()
        self .file_list .setSelectionMode (QListWidget .SelectionMode .ExtendedSelection )
        left_layout .addWidget (QLabel ("Available Files:"))
        left_layout .addWidget (self .file_list )

        self .load_button =QPushButton ("Load Selected File(s)")
        self .load_button .clicked .connect (self .load_files )
        left_layout .addWidget (self .load_button )


        self .right_panel =QTabWidget ()


        data_preview_tab =QWidget ()
        data_preview_layout =QVBoxLayout (data_preview_tab )
        self .data_table =QTableWidget ()
        data_preview_layout .addWidget (self .data_table )
        self .right_panel .addTab (data_preview_tab ,"Data Preview")


        settings_tab =QWidget ()
        settings_layout =QVBoxLayout (settings_tab )
        self .settings_table =QTableWidget ()
        settings_layout .addWidget (self .settings_table )
        self .right_panel .addTab (settings_tab ,"Settings")


        filtering_tab =QWidget ()
        filtering_layout =QVBoxLayout (filtering_tab )


        filter_group =QGroupBox ("Filter Options")
        filter_group_layout =QVBoxLayout (filter_group )

        self .filter_layout =QVBoxLayout ()
        filter_scroll_area =QScrollArea ()
        filter_scroll_area .setWidgetResizable (True )
        filter_scroll_area_content =QWidget ()
        filter_scroll_area_content .setLayout (self .filter_layout )
        filter_scroll_area .setWidget (filter_scroll_area_content )
        filter_group_layout .addWidget (filter_scroll_area )

        add_filter_button =QPushButton ("Add Filter")
        add_filter_button .clicked .connect (self .add_filter )
        filter_group_layout .addWidget (add_filter_button )

        filtering_layout .addWidget (filter_group )


        action_group =QGroupBox ("Actions")
        action_layout =QHBoxLayout (action_group )

        apply_filters_button =QPushButton ("Apply Filters")
        apply_filters_button .clicked .connect (self .apply_filters )
        action_layout .addWidget (apply_filters_button )

        save_filtered_data_button =QPushButton ("Save Filtered Data")
        save_filtered_data_button .clicked .connect (self .save_filtered_data )
        action_layout .addWidget (save_filtered_data_button )

        filtering_layout .addWidget (action_group )

        self .right_panel .addTab (filtering_tab ,"Filtering")


        comparison_tab =QWidget ()
        comparison_layout =QVBoxLayout (comparison_tab )

        self .comparison_list =QListWidget ()
        comparison_layout .addWidget (QLabel ("Loaded Files for Comparison:"))
        comparison_layout .addWidget (self .comparison_list )

        comparison_buttons_layout =QHBoxLayout ()
        self .add_comparison_file_button =QPushButton ("Add File")
        self .add_comparison_file_button .clicked .connect (self .add_comparison_file )
        comparison_buttons_layout .addWidget (self .add_comparison_file_button )

        self .remove_comparison_file_button =QPushButton ("Remove File")
        self .remove_comparison_file_button .clicked .connect (self .remove_comparison_file )
        comparison_buttons_layout .addWidget (self .remove_comparison_file_button )

        comparison_layout .addLayout (comparison_buttons_layout )

        self .compare_button =QPushButton ("Compare Files")
        self .compare_button .clicked .connect (self .compare_files )
        comparison_layout .addWidget (self .compare_button )

        self .right_panel .addTab (comparison_tab ,"Comparison")


        splitter =QSplitter (Qt .Horizontal )
        splitter .addWidget (left_panel )
        splitter .addWidget (self .right_panel )


        splitter .setSizes ([100 ,700 ])

        self .layout .addWidget (splitter )

    def setup_plotting_tab (self ):
        plotting_tab =QWidget ()
        plotting_layout =QVBoxLayout (plotting_tab )

        self .plot_options_form =QFormLayout ()

        self .plot_type_combo =QComboBox ()
        self .plot_type_combo .currentIndexChanged .connect (self .on_plot_type_changed )

        self .x_variable_combo =QComboBox ()
        self .y_variable_combo =QComboBox ()
        self .event_number_combo =QComboBox ()

        self .event_range_start =QSpinBox ()
        self .event_range_end =QSpinBox ()
        range_layout =QHBoxLayout ()
        range_layout .addWidget (QLabel ("From:"))
        range_layout .addWidget (self .event_range_start )
        range_layout .addWidget (QLabel ("To:"))
        range_layout .addWidget (self .event_range_end )

        self .center_events_combo =QComboBox ()
        self .center_events_combo .addItems (["No Centering","Center on Minimum","Center on Maximum"])

        self .same_segments_checkbox =QCheckBox ("Plot events with same number of segments")
        self .same_segments_checkbox .stateChanged .connect (self .on_same_segments_changed )

        self .segments_number_combo =QComboBox ()

        self .plot_file_combo =QComboBox ()
        self .plot_file_combo .addItem ("Current File")

        self .plot_options ={
        "plot_type":(self .plot_type_combo ,"Select Plot Type:"),
        "x_variable":(self .x_variable_combo ,"Select X Variable:"),
        "y_variable":(self .y_variable_combo ,"Select Y Variable (for scatter):"),
        "event_number":(self .event_number_combo ,"Select Event:"),
        "event_range":(range_layout ,"Event Range:"),
        "center_events":(self .center_events_combo ,"Event Centering:"),
        "same_segments":(self .same_segments_checkbox ,""),
        "segments_number":(self .segments_number_combo ,"Select number of segments:"),
        "plot_file":(self .plot_file_combo ,"Select File to Plot:")
        }

        for option ,(widget ,label )in self .plot_options .items ():
            if isinstance (widget ,QLayout ):
                self .plot_options_form .addRow (label ,widget )
            else :
                self .plot_options_form .addRow (label ,widget )


        options_group =QGroupBox ("Plot Options")
        options_group .setLayout (self .plot_options_form )
        plotting_layout .addWidget (options_group )



        button_layout =QHBoxLayout ()
        self .plot_button =QPushButton ("Generate Plot")
        self .plot_button .clicked .connect (self .plot_data )
        button_layout .addWidget (self .plot_button )
        self .plot_fitted_data_checkbox =QCheckBox ("Plot Fitted Data")
        self .plot_baseline_checkbox =QCheckBox ("Plot Baseline")
        button_layout .addWidget (self .plot_fitted_data_checkbox )
        button_layout .addWidget (self .plot_baseline_checkbox )
        plotting_layout .addLayout (button_layout )


        self .figure =plt .figure (figsize =(5 ,4 ))
        self .canvas =FigureCanvas (self .figure )
        plotting_layout .addWidget (self .canvas )

        self .right_panel .addTab (plotting_tab ,"Plotting")

        self .hide_all_plot_widgets ()
        self .update_plot_options ()



    def hide_all_plot_widgets (self ):
        if self .plot_options_form is None :
            return 

        for i in range (self .plot_options_form .rowCount ()):
            label_item =self .plot_options_form .itemAt (i ,QFormLayout .LabelRole )
            field_item =self .plot_options_form .itemAt (i ,QFormLayout .FieldRole )

            if label_item and label_item .widget ():
                label_item .widget ().setVisible (False )

            if field_item :
                if isinstance (field_item ,QLayoutItem )and field_item .widget ():
                    field_item .widget ().setVisible (False )
                elif isinstance (field_item ,QLayout ):
                    for j in range (field_item .count ()):
                        if field_item .itemAt (j )and field_item .itemAt (j ).widget ():
                            field_item .itemAt (j ).widget ().setVisible (False )

        if hasattr (self ,'plot_fitted_data_checkbox'):
            self .plot_fitted_data_checkbox .setVisible (False )
        if hasattr (self ,'plot_baseline_checkbox'):
            self .plot_baseline_checkbox .setVisible (False )




    def add_comparison_file (self ):
        file_paths ,_ =QFileDialog .getOpenFileNames (self ,"Open NPZ File(s)","","NPZ Files (*.npz)")
        for file_path in file_paths :
            try :
                data =np .load (file_path ,allow_pickle =True )
                file_name =os .path .basename (file_path )
                self .comparison_data [file_name ]=data 
                self .comparison_list .addItem (file_name )
                self .plot_file_combo .addItem (file_name )
            except Exception as e :
                QMessageBox .critical (self ,"Error",f"Failed to load file {file_name}: {str(e)}")

        if file_paths :
            QMessageBox .information (self ,"Files Added",f"Added {len(file_paths)} file(s) to comparison")

    def remove_comparison_file (self ):
        selected_items =self .comparison_list .selectedItems ()
        if selected_items :
            for item in selected_items :
                file_name =item .text ()
                del self .comparison_data [file_name ]
                self .comparison_list .takeItem (self .comparison_list .row (item ))
                index =self .plot_file_combo .findText (file_name )
                if index !=-1 :
                    self .plot_file_combo .removeItem (index )
            QMessageBox .information (self ,"Files Removed",f"Removed {len(selected_items)} file(s) from comparison")

    def compare_files (self ):
        if len (self .comparison_data )<2 :
            QMessageBox .warning (self ,"Warning","Please add at least two files for comparison.")
            return 

        self .figure .clear ()
        ax =self .figure .add_subplot (111 )

        for file_name ,data in self .comparison_data .items ():
            if 'events'in data :
                events =data ['events']
                durations =[event ['end_time']-event ['start_time']for event in events ]
                amplitudes =[np .min (event ['event_data'])for event in events ]
                ax .scatter (durations ,amplitudes ,label =file_name ,alpha =0.5 )

        ax .set_xlabel ("Duration (s)")
        ax .set_ylabel ("Amplitude")
        ax .set_title ("Event Duration vs Amplitude Comparison")

        self .canvas .draw ()

    def select_folder (self ):
        folder =QFileDialog .getExistingDirectory (self ,"Select Folder")
        if folder :
            self .current_folder =folder 
            self .update_file_list ()










    def update_file_list (self ):
        if self .current_folder :
            file_type =self .file_type_combo .currentText ()
            self .file_list .clear ()


            matching_files =[]


            for file in os .listdir (self .current_folder ):
                if file .endswith (file_type ):
                    matching_files .append (file )


            matching_files .sort (key =str .lower )


            for file in matching_files :
                self .file_list .addItem (file )

    def load_files (self ):
        selected_items =self .file_list .selectedItems ()
        if not selected_items :
            QMessageBox .warning (self ,"No File Selected","Please select one or more files to load.")
            return 

        self .comparison_data .clear ()
        self .plot_file_combo .clear ()
        self .plot_file_combo .addItem ("Current File")

        for item in selected_items :
            file_name =item .text ()
            file_path =os .path .join (self .current_folder ,file_name )
            npz_file =np .load (file_path ,allow_pickle =True )

            data ={key :npz_file [key ]for key in npz_file .files }

            if len (selected_items )==1 :
                self .loaded_data =data 
                self .original_data =copy .deepcopy (data )
                self .file_type =self .determine_file_type (file_name )
            else :
                self .comparison_data [file_name ]=data 
                self .plot_file_combo .addItem (file_name )

        if len (selected_items )==1 :
            if self .file_type =='MLdataset':
                self .update_ml_variables ()
            self .update_plot_options ()
            if self .file_type =='event_fitting':
                self .populate_segments_number_combo ()
            self .display_data ()
            self .display_settings ()
            self .update_plotting_options ()
            self .update_filtering_options ()

            if self .file_type in ['event_fitting','event_data']:
                num_events =self .get_num_events ()
                self .event_number_combo .clear ()
                self .event_number_combo .addItems ([f"Event {i}"for i in range (num_events )])

            self .right_panel .setCurrentIndex (0 )
            self .populate_segments_number_combo ()
            self .on_plot_type_changed ()
        else :
            self .comparison_list .clear ()
            self .comparison_list .addItems (self .comparison_data .keys ())
            QMessageBox .information (self ,"Files Loaded",f"Loaded {len(selected_items)} files for comparison.")


    def update_filtering_options (self ):

        for i in reversed (range (self .filter_layout .count ())):
            self .filter_layout .itemAt (i ).widget ().deleteLater ()


        filtering_tab_index =self .right_panel .indexOf (self .right_panel .findChild (QWidget ,"Filtering"))
        self .right_panel .setTabEnabled (filtering_tab_index ,self .file_type in ['event_fitting','dataset','MLdataset'])

    def update_plotting_options (self ):
        self .is_updating_plot_options =True 
        self .plot_type_combo .clear ()
        if self .file_type is None :
            self .plot_type_label .setText ("No file loaded. Please load a file to see plotting options.")
            self .hide_all_plot_widgets ()
        elif self .file_type in ['dataset','MLdataset']:
            self .plot_type_combo .addItems (["X Data Variable Distribution","X Data Variable Scatter"])
            self .update_dataset_options ()
        elif self .file_type in ['event_fitting','event_data']:
            self .plot_type_combo .addItems (["Individual Event","Multiple Events","Scatter Plot"])
            self .update_event_options ()

        self .is_updating_plot_options =False 

    def update_dataset_options (self ):
        if self .file_type =='MLdataset':
            variables =self .ml_variables 
        else :
            variables =['height','fwhm','heightatfwhm','area','width','skew','kurt','event_baseline_mean','event_time']

        self .x_variable_combo .clear ()
        self .y_variable_combo .clear ()
        self .x_variable_combo .addItems (variables )
        self .y_variable_combo .addItems (variables )



    def update_event_options (self ):
        if self .loaded_data is not None :
            num_events =self .get_num_events ()
            self .event_number_combo .clear ()
            self .event_number_combo .addItems ([f"Event {i}"for i in range (num_events )])
            self .event_range_start .setRange (0 ,max (0 ,num_events -1 ))
            self .event_range_end .setRange (0 ,max (0 ,num_events -1 ))
            self .event_range_end .setValue (min (9 ,max (0 ,num_events -1 )))



    def on_plot_type_changed (self ):
        if self .is_updating_plot_options :
            return 

        self .hide_all_plot_widgets ()
        plot_type =self .plot_type_combo .currentText ()

        self .show_plot_option ("plot_type")
        self .show_plot_option ("plot_file")
        self .plot_button .setVisible (True )

        if self .file_type in ['dataset','MLdataset']:
            self .show_plot_option ("x_variable")
            if plot_type =="X Data Variable Scatter":
                self .show_plot_option ("y_variable")
        elif self .file_type in ['event_fitting','event_data']:
            if plot_type =="Individual Event":
                self .show_plot_option ("event_number")
                self .show_plot_option ("center_events")
                if self .file_type =='event_fitting':
                    self .plot_fitted_data_checkbox .setVisible (True )
                    self .plot_baseline_checkbox .setVisible (True )
            elif plot_type =="Multiple Events":
                self .show_plot_option ("event_range")
                self .show_plot_option ("center_events")
                if self .file_type =='event_fitting':
                    self .show_plot_option ("same_segments")
                    is_checked =self .same_segments_checkbox .isChecked ()
                    if is_checked :
                        self .show_plot_option ("segments_number")
            elif plot_type =="Scatter Plot":
                self .show_plot_option ("x_variable")
                self .show_plot_option ("y_variable")

        if self .file_type =='event_fitting'and plot_type =="Multiple Events":
            self .populate_segments_number_combo ()


    def determine_file_type (self ,file_name ):
        if file_name .endswith ('.dataset.npz'):
            return 'dataset'
        elif file_name .endswith ('.event_fitting.npz'):
            return 'event_fitting'
        elif file_name .endswith ('.MLdataset.npz'):
            return 'MLdataset'
        elif file_name .endswith ('.event_data.npz'):
            return 'event_data'
        else :
            return 'unknown'


    def get_ml_scheme (self ):

        if 'settings'in self .loaded_data :

            settings =self .loaded_data ['settings']




            if isinstance (settings ,np .ndarray ):
                if settings .dtype .kind in ['U','S']:
                    settings =settings .item ()
                elif settings .dtype ==object and len (settings )>0 :
                    settings =settings [0 ]
                else :

                    return 'Unknown'


            if isinstance (settings ,str ):
                try :
                    settings_dict =json .loads (settings )

                    ml_standard =settings_dict .get ('ML_standard','Unknown')

                    return ml_standard 
                except json .JSONDecodeError as e :

                    return 'Unknown'
            else :

                return 'Unknown'
        else :

            return 'Unknown'

    def update_ml_variables (self ):
        scheme =self .get_ml_scheme ()








        if scheme =='Scheme 1':
            self .ml_variables =[f'part_{i}'for i in range (10 )]+['part_width']
        elif scheme =='Scheme 2':
            self .ml_variables =[f'part_{i}'for i in range (10 )]+['part_width','height','fwhm','heightatfwhm','area','width']
        elif scheme in ['Scheme 3','Scheme 4']:
            n_parts =10 if scheme =='Scheme 3'else 50 
            self .ml_variables =[f'part_{i}'for i in range (n_parts )]+['part_width','height','fwhm','heightatfwhm','area','width','skew','kurt']
        elif scheme =='Scheme 5':
            if 'X'in self .loaded_data and isinstance (self .loaded_data ['X'],np .ndarray ):
                signal_points =self .loaded_data ['X'].shape [1 ]-7 
                self .ml_variables =[f'signal_{i}'for i in range (signal_points )]+['height','fwhm','heightatfwhm','area','width','skew','kurt']
            else :

                self .ml_variables =['Unknown scheme']
        else :

            self .ml_variables =['Unknown scheme']



    def display_data (self ):
        if self .loaded_data is not None :
            self .clear_data_preview ()


            data_preview_tab =self .right_panel .widget (0 )
            data_preview_layout =data_preview_tab .layout ()


            self .data_table =QTableWidget ()
            self .data_table .setColumnCount (3 )
            self .data_table .setHorizontalHeaderLabels (["Key","Type","Value/Summary"])
            data_preview_layout .addWidget (self .data_table )

            if self .file_type in ['dataset','MLdataset']:
                self .display_dataset (data_preview_layout )
            elif self .file_type =='event_fitting':
                self .display_event_fitting ()
            elif self .file_type =='event_data':
                self .display_event_data ()
            else :
                self .display_generic ()

            self .data_table .resizeColumnsToContents ()


            self .right_panel .setCurrentIndex (0 )


    def clear_data_preview (self ):

        data_preview_tab =self .right_panel .widget (0 )


        if data_preview_tab .layout ()is not None :
            QWidget ().setLayout (data_preview_tab .layout ())


        QVBoxLayout (data_preview_tab )

    def display_dataset (self ,data_preview_layout ):
        main_splitter =QSplitter (Qt .Vertical )
        main_splitter .addWidget (self .data_table )


        for key ,value in self .loaded_data .items ():
            row =self .data_table .rowCount ()
            self .data_table .insertRow (row )
            self .data_table .setItem (row ,0 ,QTableWidgetItem (str (key )))
            self .data_table .setItem (row ,1 ,QTableWidgetItem (str (type (value ))))

            if key =='X'and isinstance (value ,np .ndarray ):
                self .data_table .setItem (row ,2 ,QTableWidgetItem ("See X Data details below"))
            elif key =='settings':
                self .data_table .setItem (row ,2 ,QTableWidgetItem ("See Settings tab for details"))
            else :
                self .data_table .setItem (row ,2 ,QTableWidgetItem (str (value )))


        if self .file_type =='MLdataset':
            scheme =self .get_ml_scheme ()
            row =self .data_table .rowCount ()
            self .data_table .insertRow (row )
            self .data_table .setItem (row ,0 ,QTableWidgetItem ("ML Scheme"))
            self .data_table .setItem (row ,1 ,QTableWidgetItem ("String"))
            self .data_table .setItem (row ,2 ,QTableWidgetItem (scheme ))


        if 'X'in self .loaded_data and isinstance (self .loaded_data ['X'],np .ndarray ):
            x_data_widget =QWidget ()
            x_data_layout =QVBoxLayout (x_data_widget )

            x_data_splitter =QSplitter (Qt .Vertical )


            summary_table =QTableWidget ()
            summary_table .setColumnCount (5 )
            summary_table .setHorizontalHeaderLabels (["Variable","Mean","Std","Min","Max"])
            x_data_layout .addWidget (QLabel ("X Data Summary:"))

            X_data =self .loaded_data ['X']


            if self .file_type =='MLdataset':
                variables =self .ml_variables 
            else :
                variables =['height','fwhm','heightatfwhm','area','width','skew','kurt','event_baseline_mean','event_time']

            for i ,var in enumerate (variables ):
                summary_table .insertRow (i )
                col_data =X_data [:,i ]
                summary_table .setItem (i ,0 ,QTableWidgetItem (var ))
                summary_table .setItem (i ,1 ,QTableWidgetItem (f"{col_data.mean():.8f}"))
                summary_table .setItem (i ,2 ,QTableWidgetItem (f"{col_data.std():.8f}"))
                summary_table .setItem (i ,3 ,QTableWidgetItem (f"{col_data.min():.8f}"))
                summary_table .setItem (i ,4 ,QTableWidgetItem (f"{col_data.max():.8f}"))

            summary_table .resizeColumnsToContents ()
            x_data_splitter .addWidget (summary_table )


            control_widget =QWidget ()
            control_layout =QHBoxLayout (control_widget )
            control_layout .addWidget (QLabel ("Number of rows to display:"))
            self .row_spinner =QSpinBox ()
            self .row_spinner .setRange (1 ,1000000 )
            self .row_spinner .setValue (100 )
            control_layout .addWidget (self .row_spinner )
            refresh_button =QPushButton ("Refresh")
            refresh_button .clicked .connect (self .refresh_raw_data )
            control_layout .addWidget (refresh_button )


            raw_data_widget =QWidget ()
            raw_data_layout =QVBoxLayout (raw_data_widget )
            raw_data_layout .addWidget (control_widget )
            raw_data_layout .addWidget (QLabel ("Raw X Data:"))
            self .raw_data_table =QTableWidget ()
            self .raw_data_table .setColumnCount (len (variables ))
            self .raw_data_table .setHorizontalHeaderLabels (variables )
            raw_data_layout .addWidget (self .raw_data_table )

            self .refresh_raw_data ()

            x_data_splitter .addWidget (raw_data_widget )
            x_data_layout .addWidget (x_data_splitter )

            main_splitter .addWidget (x_data_widget )


        data_preview_layout .addWidget (main_splitter )


        main_splitter .setSizes ([100 ,300 ])
        if 'X'in self .loaded_data :
            x_data_splitter .setSizes ([100 ,200 ])

    def on_same_segments_changed (self ,state ):
        is_checked =(state ==Qt .CheckState .Checked .value )
        self .plot_options ["segments_number"][0 ].setVisible (is_checked )
        if is_checked :
            self .populate_segments_number_combo ()

    def populate_segments_number_combo (self ):
        self .segments_number_combo .clear ()
        if self .loaded_data is None or self .file_type !='event_fitting':
            return 

        segment_numbers =set ()
        for key ,value in self .loaded_data .items ():
            if 'SEGMENT_INFO_'in key and 'number_of_segments'in key :
                if isinstance (value ,np .ndarray ):
                    segment_numbers .update (value .astype (int ))
                elif isinstance (value ,(int ,float )):
                    segment_numbers .add (int (value ))

        for num in sorted (segment_numbers ):
            self .segments_number_combo .addItem (str (num ))



    def display_event_fitting (self ):
        if self .loaded_data is None :
            return 


        self .data_table .setRowCount (0 )
        self .data_table .setColumnCount (4 )
        self .data_table .setHorizontalHeaderLabels (["Event","Type","Summary","Actions"])


        event_analysis_tab =QWidget ()
        event_analysis_layout =QVBoxLayout (event_analysis_tab )
        self .right_panel .addTab (event_analysis_tab ,"Event Analysis")


        self .event_analysis_table =QTableWidget ()
        self .event_analysis_table .setColumnCount (10 )
        self .event_analysis_table .setHorizontalHeaderLabels (['Event ID','height','fwhm','heightatfwhm','area','width','skew','kurt','event_baseline_mean','event_time'])
        event_analysis_layout .addWidget (self .event_analysis_table )

        event_counter =0 
        while True :
            event_key =f'EVENT_DATA_{event_counter}'
            if f'{event_key}_part_0'not in self .loaded_data :
                break 

            row =self .data_table .rowCount ()
            self .data_table .insertRow (row )
            self .data_table .setItem (row ,0 ,QTableWidgetItem (f"Event {event_counter}"))
            self .data_table .setItem (row ,1 ,QTableWidgetItem ("Event Data"))


            time_points =self .loaded_data [f'{event_key}_part_0']
            event_data =self .loaded_data [f'{event_key}_part_1']
            summary =f"Duration: {(time_points[-1] - time_points[0])*1e6:.1f}us, Points: {len(event_data)}"
            self .data_table .setItem (row ,2 ,QTableWidgetItem (summary ))


            action_widget =QWidget ()
            action_layout =QHBoxLayout (action_widget )
            view_button =QPushButton ("View")
            view_button .clicked .connect (lambda _ ,ec =event_counter :self .view_event_details (ec ))
            action_layout .addWidget (view_button )
            self .data_table .setCellWidget (row ,3 ,action_widget )


            event_analysis_key =f'EVENT_ANALYSIS_{event_counter}'
            if event_analysis_key in self .loaded_data :
                event_analysis =self .loaded_data [event_analysis_key ]
                self .event_analysis_table .insertRow (event_counter )
                self .event_analysis_table .setItem (event_counter ,0 ,QTableWidgetItem (str (event_counter )))
                for i ,value in enumerate (event_analysis ):
                    self .event_analysis_table .setItem (event_counter ,i +1 ,QTableWidgetItem (f"{value:.8f}"))

            event_counter +=1 

        self .data_table .resizeColumnsToContents ()
        self .event_analysis_table .resizeColumnsToContents ()



    def view_event_details (self ,event_counter ):
        dialog =QDialog (self )
        dialog .setWindowTitle (f"Event {event_counter} Details")
        dialog .setMinimumSize (800 ,600 )

        layout =QVBoxLayout (dialog )


        nav_layout =QHBoxLayout ()
        prev_button =QPushButton ("Previous Event")
        next_button =QPushButton ("Next Event")
        nav_layout .addWidget (prev_button )
        nav_layout .addWidget (next_button )
        layout .addLayout (nav_layout )


        tab_widget =QTabWidget ()
        layout .addWidget (tab_widget )


        event_plot_tab =QWidget ()
        event_plot_layout =QVBoxLayout (event_plot_tab )
        event_figure =plt .figure (figsize =(5 ,4 ))
        event_canvas =FigureCanvas (event_figure )
        event_plot_layout .addWidget (event_canvas )


        toolbar =NavigationToolbar (event_canvas ,event_plot_tab )
        event_plot_layout .addWidget (toolbar )

        tab_widget .addTab (event_plot_tab ,"Event Data")


        segment_info_tab =QWidget ()
        segment_info_layout =QVBoxLayout (segment_info_tab )
        segment_info_text =QTextEdit ()
        segment_info_layout .addWidget (segment_info_text )
        tab_widget .addTab (segment_info_tab ,"Segment Info")


        event_analysis_tab =QWidget ()
        event_analysis_layout =QVBoxLayout (event_analysis_tab )
        event_analysis_text =QTextEdit ()
        event_analysis_layout .addWidget (event_analysis_text )
        tab_widget .addTab (event_analysis_tab ,"Event Analysis")

        def update_event_display (ec ):
            nonlocal event_counter 
            event_counter =ec 
            dialog .setWindowTitle (f"Event {event_counter} Details")

            event_key =f'EVENT_DATA_{event_counter}'
            segment_info_key =f'SEGMENT_INFO_{event_counter}'


            event_figure .clear ()
            ax =event_figure .add_subplot (111 )
            time_points =self .loaded_data [f'{event_key}_part_0']
            event_data =self .loaded_data [f'{event_key}_part_1']
            baseline =self .loaded_data [f'{event_key}_part_4']
            fitted_data =self .loaded_data [f'{event_key}_part_3']

            ax .plot (time_points ,event_data ,label ='Event Data')
            ax .plot (time_points ,baseline ,label ='Baseline',linestyle ='--')
            ax .plot (time_points ,fitted_data ,label ='Fitted Data',linestyle ='-.')
            ax .set_title (f"Event {event_counter} Data")
            ax .set_xlabel ("Time")
            ax .set_ylabel ("Amplitude")

            event_canvas .draw ()


            segment_info =""
            for key in self .loaded_data .keys ():
                if key .startswith (segment_info_key ):
                    value =self .loaded_data [key ]
                    segment_info +=f"{key.split('_')[-1]}: {value}\n"
            segment_info_text .setText (segment_info )


            event_analysis_key =f'EVENT_ANALYSIS_{event_counter}'
            if event_analysis_key in self .loaded_data :
                event_analysis =self .loaded_data [event_analysis_key ]
                analysis_text ="Event Analysis:\n"
                for i ,value in enumerate (['height','fwhm','heightatfwhm','area','width','skew','kurt','event_baseline_mean','event_time']):
                    analysis_text +=f"{value}: {event_analysis[i]:.8f}\n"
                event_analysis_text .setText (analysis_text )


            prev_button .setEnabled (event_counter >0 )
            next_button .setEnabled (f'EVENT_DATA_{event_counter + 1}_part_0'in self .loaded_data )

        update_event_display (event_counter )

        def go_to_previous_event ():
            if event_counter >0 :
                update_event_display (event_counter -1 )

        def go_to_next_event ():
            if f'EVENT_DATA_{event_counter + 1}_part_0'in self .loaded_data :
                update_event_display (event_counter +1 )

        prev_button .clicked .connect (go_to_previous_event )
        next_button .clicked .connect (go_to_next_event )

        dialog .exec ()

    def refresh_raw_data (self ):
        if 'X'in self .loaded_data and isinstance (self .loaded_data ['X'],np .ndarray ):
            X_data =self .loaded_data ['X']
            requested_rows =self .row_spinner .value ()
            num_rows =min (requested_rows ,X_data .shape [0 ])

            self .raw_data_table .setRowCount (0 )
            for i in range (num_rows ):
                self .raw_data_table .insertRow (i )
                for j in range (X_data .shape [1 ]):
                    self .raw_data_table .setItem (i ,j ,QTableWidgetItem (f"{X_data[i, j]:.8f}"))

            self .raw_data_table .resizeColumnsToContents ()

            if num_rows <requested_rows :
                QMessageBox .information (self ,"Row Limit",f"Only {num_rows} rows available. Displaying all available rows.")

    def summarize_X_data (self ,X ):
        variables =['height','fwhm','heightatfwhm','area','width','skew','kurt','event_baseline_mean','event_time']
        summary ="X data summary:\n"
        for i ,var in enumerate (variables ):
            col_data =X [:,i ]
            summary +=f"{var}: Mean={col_data.mean():.8f}, Std={col_data.std():.8f}, Min={col_data.min():.8f}, Max={col_data.max():.8f}\n"
        return summary 

    def display_event_data (self ):
        if 'events'in self .loaded_data :
            events =self .loaded_data ['events']
            for i ,event in enumerate (events ):
                row =self .data_table .rowCount ()
                self .data_table .insertRow (row )
                self .data_table .setItem (row ,0 ,QTableWidgetItem (f"Event {i}"))
                self .data_table .setItem (row ,1 ,QTableWidgetItem ("Event Data"))
                summary =f"Start: {event['start_time']:.8f}, End: {event['end_time']:.8f}, Duration: {event['end_time'] - event['start_time']:.8f}"
                self .data_table .setItem (row ,2 ,QTableWidgetItem (summary ))

    def display_generic (self ):
        for key ,value in self .loaded_data .items ():
            row =self .data_table .rowCount ()
            self .data_table .insertRow (row )
            self .data_table .setItem (row ,0 ,QTableWidgetItem (str (key )))
            self .data_table .setItem (row ,1 ,QTableWidgetItem (str (type (value ))))
            if isinstance (value ,np .ndarray ):
                summary =f"Shape: {value.shape}"
                if value .dtype .kind in ['i','f']:
                    summary +=f", Mean: {value.mean():.2f}, Std: {value.std():.2f}"
                self .data_table .setItem (row ,2 ,QTableWidgetItem (summary ))
            else :
                self .data_table .setItem (row ,2 ,QTableWidgetItem (str (value )))



    def display_settings (self ):
        try :
            if self .loaded_data is not None and 'settings'in self .loaded_data :
                settings =self .loaded_data ['settings']


                if isinstance (settings ,np .ndarray ):
                    if settings .dtype ==object and len (settings )>0 :
                        settings =settings [0 ]
                    else :
                        settings =str (settings )

                try :

                    if isinstance (settings ,str ):
                        settings_dict =json .loads (settings )
                    elif isinstance (settings ,dict ):
                        settings_dict =settings 
                    else :
                        raise ValueError ("Settings data is not in a recognized format")

                    self .settings_table .clear ()
                    self .settings_table .setRowCount (0 )
                    self .settings_table .setColumnCount (2 )
                    self .settings_table .setHorizontalHeaderLabels (["Setting","Value"])

                    for key ,value in settings_dict .items ():
                        row =self .settings_table .rowCount ()
                        self .settings_table .insertRow (row )
                        self .settings_table .setItem (row ,0 ,QTableWidgetItem (str (key )))
                        self .settings_table .setItem (row ,1 ,QTableWidgetItem (str (value )))

                    self .settings_table .resizeColumnsToContents ()
                except json .JSONDecodeError :
                    print ("Error decoding settings JSON")
                    self .display_settings_as_string (settings )
                except ValueError as e :
                    print (f"Error processing settings: {str(e)}")
                    self .display_settings_as_string (settings )
            else :
                self .settings_table .clear ()
                self .settings_table .setRowCount (1 )
                self .settings_table .setColumnCount (1 )
                self .settings_table .setItem (0 ,0 ,QTableWidgetItem ("No settings data available"))
        except :
            pass 

    def display_settings_as_string (self ,settings ):
        self .settings_table .clear ()
        self .settings_table .setRowCount (1 )
        self .settings_table .setColumnCount (1 )
        self .settings_table .setItem (0 ,0 ,QTableWidgetItem (str (settings )))
        self .settings_table .resizeColumnsToContents ()

    def update_plot_options (self ):
        self .is_updating_plot_options =True 
        self .hide_all_plot_widgets ()

        self .plot_type_combo .clear ()
        self .show_plot_option ("plot_type")
        self .show_plot_option ("plot_file")
        self .plot_button .setVisible (True )

        if self .file_type is None :
            self .plot_type_combo .addItem ("No file loaded")
            self .plot_type_combo .setEnabled (False )
        else :
            self .plot_type_combo .setEnabled (True )
            if self .file_type in ['dataset','MLdataset']:
                self .plot_type_combo .addItems (["X Data Variable Distribution","X Data Variable Scatter"])
                self .update_dataset_options ()
            elif self .file_type in ['event_fitting','event_data']:
                self .plot_type_combo .addItems (["Individual Event","Multiple Events","Event Properties Histogram","Scatter Plot"])
                self .update_event_options ()

        self .on_plot_type_changed ()
        self .is_updating_plot_options =False 

    def show_plot_option (self ,option ):
        if self .plot_options_form is None :
            return 

        index =list (self .plot_options .keys ()).index (option )
        label_item =self .plot_options_form .itemAt (index ,QFormLayout .LabelRole )
        field_item =self .plot_options_form .itemAt (index ,QFormLayout .FieldRole )

        if label_item and label_item .widget ():
            label_item .widget ().setVisible (True )

        if field_item :
            if isinstance (field_item ,QLayoutItem )and field_item .widget ():
                field_item .widget ().setVisible (True )
            elif isinstance (field_item ,QLayout ):
                for i in range (field_item .count ()):
                    if field_item .itemAt (i )and field_item .itemAt (i ).widget ():
                        field_item .itemAt (i ).widget ().setVisible (True )


    def get_num_events (self ):
        if self .file_type =='event_fitting':

            event_numbers =set ()
            for key in self .loaded_data .keys ():
                if key .startswith ('EVENT_DATA_'):
                    parts =key .split ('_')
                    if len (parts )>2 and parts [2 ].isdigit ():
                        event_numbers .add (int (parts [2 ]))
            return len (event_numbers )
        elif self .file_type =='event_data'and 'events'in self .loaded_data :
            return len (self .loaded_data ['events'])
        return 0 

    def plot_data (self ):
        if self .loaded_data is None and not self .comparison_data :
            print ("No data loaded.")
            return 

        plot_type =self .plot_type_combo .currentText ()
        selected_file =self .plot_file_combo .currentText ()

        self .figure .clear ()
        ax =self .figure .add_subplot (111 )

        if selected_file =="Current File":
            data_to_plot ={"Current File":self .loaded_data }if self .loaded_data else self .comparison_data 
        else :
            data_to_plot ={selected_file :self .comparison_data [selected_file ]}

        try :
            if self .file_type in ['dataset','MLdataset']:
                self .plot_dataset_data (ax ,data_to_plot ,plot_type )
            elif self .file_type in ['event_fitting','event_data']:
                self .plot_event_data (ax ,data_to_plot ,plot_type )

            if plot_type !="Multiple Events"and ax .get_legend_handles_labels ()[0 ]:
                ax .legend ()

            self .canvas .draw ()
        except Exception as e :
            print (f"Error generating plot: {str(e)}")
            import traceback 
            traceback .print_exc ()

        QApplication .processEvents ()

    def plot_dataset_data (self ,ax ,data_to_plot ,plot_type ):
        if plot_type =="X Data Variable Distribution":
            self .plot_x_data_variable (ax ,data_to_plot )
        elif plot_type =="X Data Variable Scatter":
            self .plot_x_data_scatter (ax ,data_to_plot )

    def plot_event_data (self ,ax ,data_to_plot ,plot_type ):
        if plot_type =="Individual Event":
            centering =self .center_events_combo .currentText ()
            self .plot_single_event (ax ,data_to_plot ,centering )
        elif plot_type =="Multiple Events":
            centering =self .center_events_combo .currentText ()
            self .plot_multiple_events (ax ,data_to_plot ,centering )
        elif plot_type =="Event Properties Histogram":
            self .plot_event_properties_histogram (ax ,data_to_plot )
        elif plot_type =="Scatter Plot":
            self .plot_scatter (ax ,data_to_plot )



    def plot_x_data_distribution (self ,ax ):
        variable =self .x_variable_combo .currentText ()
        var_index =['height','fwhm','heightatfwhm','area','width','skew','kurt','event_baseline_mean','event_time'].index (variable )
        X_data =self .loaded_data ['X']
        variable_data =X_data [:,var_index ]
        ax .hist (variable_data ,bins =30 )
        ax .set_title (f"{variable} Distribution")
        ax .set_xlabel (variable )
        ax .set_ylabel ("Frequency")

    def plot_x_data_scatter (self ,ax ,data_to_plot ):
        x_variable =self .x_variable_combo .currentText ()
        y_variable =self .y_variable_combo .currentText ()

        if self .file_type =='MLdataset':
            x_index =self .ml_variables .index (x_variable )
            y_index =self .ml_variables .index (y_variable )
        else :
            variables =['height','fwhm','heightatfwhm','area','width','skew','kurt','event_baseline_mean','event_time']
            x_index =variables .index (x_variable )
            y_index =variables .index (y_variable )

        for label ,data in data_to_plot .items ():
            if 'X'in data :
                X_data =data ['X']
                ax .scatter (X_data [:,x_index ],X_data [:,y_index ],label =label ,alpha =0.5 )

        ax .set_title (f"{y_variable} vs {x_variable}")
        ax .set_xlabel (x_variable )
        ax .set_ylabel (y_variable )


    def plot_single_event (self ,ax ,data_to_plot ,centering ):
        event_number =int (self .event_number_combo .currentText ().split ()[-1 ])
        for label ,data in data_to_plot .items ():
            if self .file_type =='event_fitting':
                event_key =f'EVENT_DATA_{event_number}'
                if f'{event_key}_part_0'not in data :
                    continue 
                time_points =data [f'{event_key}_part_0']
                event_data =data [f'{event_key}_part_1']
                baseline =data [f'{event_key}_part_4']
                fitted_data =data [f'{event_key}_part_3']

                ax .plot (time_points ,event_data ,label =f'{label} Event {event_number}')

                if self .plot_fitted_data_checkbox .isChecked ():
                    ax .plot (time_points ,fitted_data ,label =f'{label} Fitted Data',linestyle ='--')

                if self .plot_baseline_checkbox .isChecked ():
                    ax .plot (time_points ,baseline ,label =f'{label} Baseline',linestyle =':')

            elif 'events'in data :
                events =data ['events']
                if event_number >=len (events ):
                    continue 
                event =events [event_number ]
                time_points =np .arange (len (event ['event_data']))/data .get ('sampling_rate',1 )*1000 
                event_data =event ['event_data']
                ax .plot (time_points ,event_data ,label =f'{label} Event {event_number}')

        ax .set_title (f"Event {event_number} Data")
        ax .set_xlabel ("Time (ms)")
        ax .set_ylabel ("Amplitude")

    def plot_multiple_events (self ,ax ,data_to_plot ,centering ):
        start =self .event_range_start .value ()
        end =self .event_range_end .value ()
        colormap =plt .get_cmap ('viridis')

        plot_same_segments =self .same_segments_checkbox .isChecked ()
        selected_segments =None 
        if plot_same_segments and self .segments_number_combo .currentText ():
            try :
                selected_segments =int (self .segments_number_combo .currentText ())
            except ValueError :
                print ("Invalid segment number selected")

        for label ,data in data_to_plot .items ():
            events_to_plot =[]
            if self .file_type =='event_fitting':
                for event_number in range (start ,end +1 ):
                    event_key =f'EVENT_DATA_{event_number}'
                    if f'{event_key}_part_0'not in data :
                        continue 
                    time_points =data [f'{event_key}_part_0']
                    event_data =data [f'{event_key}_part_1']
                    baseline =data [f'{event_key}_part_4'][0 ]
                    segment_info_key =f'SEGMENT_INFO_{event_number}_number_of_segments'
                    num_segments =data .get (segment_info_key ,[0 ])[0 ]
                    if not plot_same_segments or (plot_same_segments and num_segments ==selected_segments ):
                        events_to_plot .append ((time_points ,event_data ,baseline ,num_segments ))
            elif 'events'in data :
                events =data ['events'][start :end +1 ]
                events_to_plot =[
                (np .arange (len (event ['event_data']))/data .get ('sampling_rate',1 )*1000 ,
                event ['event_data'],
                np .mean (event ['event_data'][:1 ]),
                event .get ('num_segments',0 ))
                for event in events 
                if not plot_same_segments or (plot_same_segments and event .get ('num_segments',0 )==selected_segments )
                ]

            for i ,(time_points ,event_data ,baseline ,num_segments )in enumerate (events_to_plot ):
                centered_time ,centered_data =self .center_event (time_points ,event_data ,centering )
                color =colormap (i /len (events_to_plot )if events_to_plot else 0 )
                ax .plot (centered_time ,centered_data -baseline ,color =color ,alpha =0.7 ,
                label =f'{label} Event {start+i} (Segments: {num_segments})')

        ax .set_title (f"Multiple Events{' (Same Segments)' if plot_same_segments else ''}")
        ax .set_xlabel ("Time (ms)")
        ax .set_ylabel ("Amplitude")
        if len (events_to_plot )<0 :
            ax .text (0.5 ,0.5 ,"No events match the current criteria",
            ha ='center',va ='center',transform =ax .transAxes )


    def center_event (self ,time_points ,event_data ,centering ):
        if centering =="Center on Minimum":
            center_index =np .argmin (event_data )
        elif centering =="Center on Maximum":
            center_index =np .argmax (event_data )
        else :
            center_index =len (event_data )//2 

        centered_time =time_points -time_points [center_index ]
        return centered_time ,event_data 

    def plot_segment_mean_differences (self ,ax ):
        event_number =int (self .event_number_combo .currentText ().split ()[-1 ])
        segment_mean_diffs =self .loaded_data [f'SEGMENT_INFO_{event_number}_segment_mean_diffs']
        ax .plot (segment_mean_diffs )
        ax .set_title (f"Event {event_number} Segment Mean Differences")
        ax .set_xlabel ("Segment")
        ax .set_ylabel ("Mean Difference")

    def plot_segment_widths (self ,ax ):
        event_number =int (self .event_number_combo .currentText ().split ()[-1 ])
        segment_widths =self .loaded_data [f'SEGMENT_INFO_{event_number}_segment_widths_time']
        ax .plot (segment_widths )
        ax .set_title (f"Event {event_number} Segment Widths")
        ax .set_xlabel ("Segment")
        ax .set_ylabel ("Width (time)")

    def plot_all_events (self ,ax ):
        events =self .loaded_data ['events']
        for event in events :
            event_data =event ['event_data']
            time_points =np .linspace (event ['start_time'],event ['end_time'],len (event_data ))
            ax .plot (time_points ,event_data )
        ax .set_title ("All Events")
        ax .set_xlabel ("Time")
        ax .set_ylabel ("Amplitude")

    def plot_event_duration_distribution (self ,ax ):
        events =self .loaded_data ['events']
        durations =[event ['end_time']-event ['start_time']for event in events ]
        ax .hist (durations ,bins =30 )
        ax .set_title ("Event Duration Distribution")
        ax .set_xlabel ("Duration")
        ax .set_ylabel ("Frequency")

    def plot_x_data_variable (self ,ax ,data_to_plot ):
        variable =self .x_variable_combo .currentText ()
        for label ,data in data_to_plot .items ():
            if 'X'in data :
                X_data =data ['X']
                if self .file_type =='MLdataset':
                    var_index =self .ml_variables .index (variable )
                else :
                    var_index =['height','fwhm','heightatfwhm','area','width','skew','kurt','event_baseline_mean','event_time'].index (variable )
                variable_data =X_data [:,var_index ]
                ax .hist (variable_data ,bins =30 ,alpha =0.5 ,label =label )

        ax .set_title (f"{variable} Distribution")
        ax .set_xlabel (variable )
        ax .set_ylabel ("Frequency")


    def plot_event_properties_histogram (self ,ax ,data_to_plot ):
        for label ,data in data_to_plot .items ():
            if 'events'in data :
                events =data ['events']
                durations =[event ['end_time']-event ['start_time']for event in events ]
                amplitudes =[np .min (event ['event_data'])for event in events ]
                ax .hist (durations ,bins =30 ,alpha =0.5 ,label =f'{label} Duration')
                ax .hist (amplitudes ,bins =30 ,alpha =0.5 ,label =f'{label} Amplitude')

        ax .set_title ("Event Properties Histogram")
        ax .set_xlabel ("Value")
        ax .set_ylabel ("Frequency")


    def plot_scatter (self ,ax ,data_to_plot ):
        for label ,data in data_to_plot .items ():
            if 'events'in data :
                events =data ['events']
                durations =[event ['end_time']-event ['start_time']for event in events ]
                amplitudes =[np .min (event ['event_data'])for event in events ]
                ax .scatter (durations ,amplitudes ,label =label ,alpha =0.5 )

        ax .set_title ("Event Duration vs Amplitude")
        ax .set_xlabel ("Duration (s)")
        ax .set_ylabel ("Amplitude")



    def add_filter (self ):
        filter_widget =QWidget ()
        filter_layout =QGridLayout (filter_widget )

        variable_label =QLabel ("Variable:")
        variable_combo_box =QComboBox ()
        variables =self .get_filter_variables ()
        variable_combo_box .addItems (variables )
        filter_layout .addWidget (variable_label ,0 ,0 )
        filter_layout .addWidget (variable_combo_box ,0 ,1 )

        min_value_label =QLabel ("Min Value:")
        min_value_spin_box =QDoubleSpinBox ()
        min_value_spin_box .setDecimals (7 )
        min_value_spin_box .setRange (-1e20 ,1e20 )
        min_value_spin_box .setStepType (QDoubleSpinBox .StepType .AdaptiveDecimalStepType )
        filter_layout .addWidget (min_value_label ,1 ,0 )
        filter_layout .addWidget (min_value_spin_box ,1 ,1 )

        max_value_label =QLabel ("Max Value:")
        max_value_spin_box =QDoubleSpinBox ()
        max_value_spin_box .setDecimals (7 )
        max_value_spin_box .setRange (-1e20 ,1e20 )
        max_value_spin_box .setStepType (QDoubleSpinBox .StepType .AdaptiveDecimalStepType )
        filter_layout .addWidget (max_value_label ,2 ,0 )
        filter_layout .addWidget (max_value_spin_box ,2 ,1 )

        remove_filter_button =QPushButton ("Remove")
        remove_filter_button .clicked .connect (lambda :self .remove_filter (filter_widget ))
        filter_layout .addWidget (remove_filter_button ,3 ,0 ,1 ,2 )

        self .filter_layout .addWidget (filter_widget )
        self .update_logic_combo_boxes ()

    def get_filter_variables (self ):
        if self .file_type =='event_fitting':
            return ['height','fwhm','heightatfwhm','area','width','skew','kurt','event_baseline_mean','event_time']
        elif self .file_type =='MLdataset':
            return self .ml_variables 
        elif self .file_type =='dataset':
            return ['height','fwhm','heightatfwhm','area','width','skew','kurt','event_baseline_mean','event_time']
        else :
            return []

    def update_logic_combo_boxes (self ):
        filter_count =self .filter_layout .count ()

        for i in range (filter_count -1 ):
            filter_widget =self .filter_layout .itemAt (i ).widget ()
            filter_layout =filter_widget .layout ()

            if filter_layout .rowCount ()==4 :
                logic_combo_box =QComboBox ()
                logic_combo_box .addItems (["AND","OR"])
                filter_layout .addWidget (QLabel ("Logic:"),4 ,0 )
                filter_layout .addWidget (logic_combo_box ,4 ,1 )
            elif i <filter_count -2 :
                logic_label =filter_layout .itemAtPosition (4 ,0 )
                if logic_label is None :
                    logic_combo_box =QComboBox ()
                    logic_combo_box .addItems (["AND","OR"])
                    filter_layout .addWidget (QLabel ("Logic:"),4 ,0 )
                    filter_layout .addWidget (logic_combo_box ,4 ,1 )
            else :
                logic_label =filter_layout .itemAtPosition (4 ,0 )
                if logic_label :
                    filter_layout .removeItem (filter_layout .itemAtPosition (4 ,0 ))
                    filter_layout .removeItem (filter_layout .itemAtPosition (4 ,1 ))
                    logic_label .widget ().deleteLater ()
                    filter_layout .itemAtPosition (4 ,1 ).widget ().deleteLater ()

    def remove_filter (self ,filter_widget ):
        self .filter_layout .removeWidget (filter_widget )
        filter_widget .deleteLater ()
        self .update_logic_combo_boxes ()

    def apply_filters (self ):
        if self .loaded_data is None :
            QMessageBox .warning (self ,"Warning","No data loaded to filter.")
            return 

        mask =np .ones (self .get_data_length (),dtype =bool )

        for i in range (self .filter_layout .count ()):
            filter_widget =self .filter_layout .itemAt (i ).widget ()
            filter_layout =filter_widget .layout ()

            variable =filter_layout .itemAtPosition (0 ,1 ).widget ().currentText ()
            min_value =filter_layout .itemAtPosition (1 ,1 ).widget ().value ()
            max_value =filter_layout .itemAtPosition (2 ,1 ).widget ().value ()

            try :
                variable_values =self .get_variable_values (variable )
                condition =(variable_values >=min_value )&(variable_values <=max_value )

                if i ==0 :
                    mask =condition 
                else :
                    logic_combo =filter_layout .itemAtPosition (4 ,1 )
                    if logic_combo :
                        logic =logic_combo .widget ().currentText ()
                        if logic =="AND":
                            mask &=condition 
                        else :
                            mask |=condition 
                    else :
                        mask &=condition 
            except ValueError as e :
                QMessageBox .warning (self ,"Filter Error",f"Error applying filter for variable '{variable}': {str(e)}")
                return 

        self .apply_mask (mask )
        self .display_data ()
        QMessageBox .information (self ,"Filter Applied",f"Filtered data. {np.sum(mask)} items remaining.")

    def get_data_length (self ):
        if self .file_type =='event_fitting':
            return len ([key for key in self .loaded_data .keys ()if key .startswith ('EVENT_DATA_')])
        elif self .file_type in ['dataset','MLdataset']:
            return len (self .loaded_data ['X'])
        else :
            return 0 

    def get_variable_values (self ,variable ):
        if self .file_type =='event_fitting':
            return np .array ([self .loaded_data [f'EVENT_ANALYSIS_{i}'][self .get_variable_index (variable )]
            for i in range (self .get_data_length ())])
        elif self .file_type in ['dataset','MLdataset']:
            return self .loaded_data ['X'][:,self .get_variable_index (variable )]

    def get_variable_index (self ,variable ):
        if self .file_type =='MLdataset':
            return self .ml_variables .index (variable )
        elif self .file_type in ['dataset','event_fitting']:
            variables =['height','fwhm','heightatfwhm','area','width','skew','kurt','event_baseline_mean','event_time']
            return variables .index (variable )
        else :
            raise ValueError (f"Unsupported file type: {self.file_type}")

    def apply_mask (self ,mask ):
        if self .file_type =='event_fitting':
            self .loaded_data ={k :v for k ,v in self .loaded_data .items ()
            if not k .startswith ('EVENT_')or int (k .split ('_')[2 ])in np .where (mask )[0 ]}
        elif self .file_type in ['dataset','MLdataset']:
            self .loaded_data ['X']=self .loaded_data ['X'][mask ]

    def save_filtered_data (self ):
        if self .loaded_data is None :
            QMessageBox .warning (self ,"Warning","No data loaded to save.")
            return 

        file_path ,_ =QFileDialog .getSaveFileName (self ,"Save Filtered Data","","NPZ Files (*.npz)")
        if file_path :
            try :
                np .savez_compressed (file_path ,**self .loaded_data )
                QMessageBox .information (self ,"Save Successful",f"Filtered data saved to {file_path}")
            except Exception as e :
                QMessageBox .critical (self ,"Save Failed",f"Failed to save filtered data: {str(e)}")

    def reset_filter (self ):
        if hasattr (self ,'original_data'):
            self .loaded_data =self .original_data .copy ()
            self .display_data ()
            QMessageBox .information (self ,"Filter Reset","Data reset to original state.")
        else :
            QMessageBox .warning (self ,"Warning","No original data available to reset to.")

    def export_to_npz (self ):
        if self .loaded_data is None :
            QMessageBox .warning (self ,"Warning","No data loaded to export.")
            return 

        file_path ,_ =QFileDialog .getSaveFileName (self ,"Save NPZ File","","NPZ Files (*.npz)")
        if file_path :
            try :
                np .savez_compressed (file_path ,**self .loaded_data )
                QMessageBox .information (self ,"Export Successful",f"Data exported to {file_path}")
            except Exception as e :
                QMessageBox .critical (self ,"Export Failed",f"Failed to export data: {str(e)}")

    def export_to_csv (self ):
        if self .loaded_data is None or 'events'not in self .loaded_data :
            QMessageBox .warning (self ,"Warning","No event data loaded to export.")
            return 

        file_path ,_ =QFileDialog .getSaveFileName (self ,"Save CSV File","","CSV Files (*.csv)")
        if file_path :
            try :
                events =self .loaded_data ['events']
                with open (file_path ,'w',newline ='')as csvfile :
                    writer =csv .writer (csvfile )
                    writer .writerow (['Event ID','Start Time','End Time','Duration','Min Amplitude','Max Amplitude'])

                    for i ,event in enumerate (events ):
                        start_time =event ['start_time']
                        end_time =event ['end_time']
                        duration =end_time -start_time 
                        min_amplitude =np .min (event ['event_data'])
                        max_amplitude =np .max (event ['event_data'])

                        writer .writerow ([i ,start_time ,end_time ,duration ,min_amplitude ,max_amplitude ])

                QMessageBox .information (self ,"Export Successful",f"Event data exported to {file_path}")
            except Exception as e :
                QMessageBox .critical (self ,"Export Failed",f"Failed to export data: {str(e)}")




if __name__ =="__main__":
    app =QApplication (sys .argv )
    app .setStyle (QStyleFactory .create ('Fusion'))

    palette =QPalette ()
    palette .setColor (QPalette .ColorRole .Window ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .WindowText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Base ,QColor (25 ,25 ,25 ))
    palette .setColor (QPalette .ColorRole .AlternateBase ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ToolTipBase ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .ToolTipText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Text ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Button ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ButtonText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .BrightText ,Qt .GlobalColor .red )
    palette .setColor (QPalette .ColorRole .Link ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .Highlight ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .HighlightedText ,Qt .GlobalColor .black )
    app .setPalette (palette )
    window =SDDatabaseViewer ()
    window .showMaximized ()
    sys .exit (app .exec ())


from __future__ import absolute_import ,print_function 

import os 
import sys 

from macholib import macho_dump ,macho_standalone 
from macholib .util import is_platform_file 

gCommand =None 


def check_file (fp ,path ,callback ):
    if not os .path .exists (path ):
        print ("%s: %s: No such file or directory"%(gCommand ,path ),file =sys .stderr )
        return 1 

    try :
        is_plat =is_platform_file (path )

    except IOError as msg :
        print ("%s: %s: %s"%(gCommand ,path ,msg ),file =sys .stderr )
        return 1 

    else :
        if is_plat :
            callback (fp ,path )
    return 0 


def walk_tree (callback ,paths ):
    err =0 

    for base in paths :
        if os .path .isdir (base ):
            for root ,_dirs ,files in os .walk (base ):
                for fn in files :
                    err |=check_file (sys .stdout ,os .path .join (root ,fn ),callback )
        else :
            err |=check_file (sys .stdout ,base ,callback )

    return err 


def print_usage (fp ):
    print ("Usage:",file =fp )
    print ("  python -mmacholib [help|--help]",file =fp )
    print ("  python -mmacholib dump FILE ...",file =fp )
    print ("  python -mmacholib find DIR ...",file =fp )
    print ("  python -mmacholib standalone DIR ...",file =fp )


def main ():
    global gCommand 
    if len (sys .argv )<3 :
        print_usage (sys .stderr )
        sys .exit (1 )

    gCommand =sys .argv [1 ]

    if gCommand =="dump":
        walk_tree (macho_dump .print_file ,sys .argv [2 :])

    elif gCommand =="find":
        walk_tree (lambda fp ,path :print (path ,file =fp ),sys .argv [2 :])

    elif gCommand =="standalone":
        for dn in sys .argv [2 :]:
            macho_standalone .standaloneApp (dn )

    elif gCommand in ("help","--help"):
        print_usage (sys .stdout )
        sys .exit (0 )

    else :
        print_usage (sys .stderr )
        sys .exit (1 )


if __name__ =="__main__":
    main ()



import os 
import numpy as np 
from PySide6 .QtWidgets import (QApplication ,QWidget ,QHBoxLayout ,QVBoxLayout ,QLabel ,
QPushButton ,QCheckBox ,QListWidget ,QFileDialog ,QTabWidget ,
QComboBox ,QDoubleSpinBox ,QGroupBox ,QFormLayout ,QLineEdit ,
QTableWidget ,QTableWidgetItem ,QScrollArea ,QSpinBox ,QRadioButton ,QListWidgetItem ,QStyleFactory ,QMessageBox ,
QGridLayout )
from PySide6 .QtCore import Qt 
from PySide6 .QtGui import QPalette ,QColor ,QFont 
from matplotlib .backends .backend_qt5agg import FigureCanvasQTAgg as FigureCanvas 
from matplotlib .backends .backend_qt5agg import NavigationToolbar2QT as NavigationToolbar 
import matplotlib .pyplot as plt 
from sklearn .cluster import KMeans ,AgglomerativeClustering ,AffinityPropagation ,MeanShift 
from sklearn .mixture import GaussianMixture 
from sklearn .metrics import silhouette_score ,davies_bouldin_score 
from joblib import Parallel ,delayed 
from tqdm import tqdm 
import multiprocessing as mp 
import math 
import scipy .stats as stats 
import scipy .signal as sig 
from datetime import datetime 
from scipy .ndimage import uniform_filter1d 



def calculate_delta_I1 (l ,sigma ,V ,i0 ,i01 ,delta_I ):
    d =(i0 +np .sqrt (i0 *(i0 +(16 *l *V *sigma )/math .pi )))/(2 *V *sigma )
    d1 =(i01 +np .sqrt (i01 *(i01 +(16 *l *V *sigma )/math .pi )))/(2 *V *sigma )
    delta_G =delta_I /V 

    term1 =(d1 **2 *math .pi *sigma )/(4 *l +d1 *math .pi )

    sqrt_term =(4 *l *delta_G +d *math .pi *(delta_G -d *sigma ))**3 *(-64 *l **2 *sigma +4 *l *math .pi *(delta_G -4 *d *sigma )+d *math .pi **2 *(delta_G -d *sigma ))
    sqrt_term =np .where (sqrt_term >=0 ,sqrt_term ,np .nan )

    term2_numerator =(
    -d **2 *math .pi **3 *delta_G **2 +
    128 *l **3 *delta_G *sigma +
    2 *d **3 *math .pi **3 *delta_G *sigma +
    d **4 *math .pi **3 *sigma **2 -
    2 *d **2 *d1 **2 *math .pi **3 *sigma **2 -
    16 *l **2 *math .pi *(delta_G **2 -4 *d *delta_G *sigma +2 *d1 **2 *sigma **2 )+
    8 *d *l *math .pi **2 *(-delta_G **2 +2 *d *delta_G *sigma +(d **2 -2 *d1 **2 )*sigma **2 )-
    np .sqrt (math .pi )*np .sqrt (sqrt_term )
    )

    term2_denominator =(
    (4 *l +d *math .pi )**2 *sigma *(
    8 *l +np .sqrt (2 *math .pi )*np .sqrt (
    1 /((4 *l +d *math .pi )**2 *sigma **2 )*(
    d **2 *math .pi **3 *delta_G **2 -
    128 *l **3 *delta_G *sigma -
    2 *d **3 *math .pi **3 *delta_G *sigma -
    d **4 *math .pi **3 *sigma **2 +
    2 *d **2 *d1 **2 *math .pi **3 *sigma **2 +
    16 *l **2 *math .pi *(delta_G **2 -4 *d *delta_G *sigma +2 *d1 **2 *sigma **2 )-
    8 *d *l *math .pi **2 *(-delta_G **2 +2 *d *delta_G *sigma +(d **2 -2 *d1 **2 )*sigma **2 )+
    np .sqrt (math .pi )*np .sqrt (sqrt_term )
    )
    )
    )
    )

    term2 =term2_numerator /term2_denominator 

    delta_I1 =V *(term1 +term2 )

    return np .where (np .isnan (delta_I1 ),np .nan ,delta_I1 )



def normalize_signal (signal ,standard_length_nm ,standard_conductivity_S_m ,standard_voltage_applied_mV ,standard_open_pore_current_nA ,event_baseline_mean ):
    normalized_signal =calculate_delta_I1 (
    standard_length_nm *10 **(-9 ),
    standard_conductivity_S_m ,
    standard_voltage_applied_mV /1000 ,
    standard_open_pore_current_nA *10 **(-9 ),
    event_baseline_mean *10 **(-9 ),
    signal *10 **(-9 )
    )
    return normalized_signal /10 **(-9 )


def standardize_events (events ,standard ,event_baseline_mean ,standard_power ,standard_length_nm ,standard_conductivity_S_m ,standard_voltage_applied_mV ,standard_open_pore_current_nA ):
    standardized_events =[]

    for event in events :
        if standard =="Normal":
            standardized_event =event 
        elif standard =="ΔI/I₀":
            standardized_event =event /event_baseline_mean 
        elif standard =="(ΔI*I₀)⁰·⁵":
            standardized_event =(np .abs (event )*np .abs (event_baseline_mean ))**0.5 
        elif standard =="(ΔI*I₀)ᵖᵒʷᵉʳ":
            standardized_event =(np .abs (event )*np .abs (event_baseline_mean ))**standard_power 
        elif standard =="Dutt Standardisation":
            standardized_event =normalize_signal (event ,standard_length_nm ,standard_conductivity_S_m ,standard_voltage_applied_mV ,standard_open_pore_current_nA ,event_baseline_mean )
        else :
            standardized_event =event 

        standardized_events .append (standardized_event )

    return standardized_events 

def analyze_event (event_signal ,event_baseline_mean ,sampling_rate ,ML_standardisation_settings ):
    standard ,ML_enabled ,ML_standard ,standard_power ,standard_length_nm ,standard_conductivity_S_m ,standard_voltage_applied_mV ,standard_open_pore_current_nA =ML_standardisation_settings ['standard'],ML_standardisation_settings ['ML_enabled'],ML_standardisation_settings ['ML_standard'],ML_standardisation_settings ['standard_power'],ML_standardisation_settings ['standard_length_nm'],ML_standardisation_settings ['standard_conductivity_S_m'],ML_standardisation_settings ['standard_voltage_applied_mV'],ML_standardisation_settings ['standard_open_pore_current_nA']

    if ML_enabled =="False":
        signal =np .abs (event_signal )

        sample_period =1 /sampling_rate 
        width =len (signal )*sample_period 
        area =np .sum (signal )*sample_period 
        peaks ,_ =sig .find_peaks (signal )
        if len (peaks )==0 :
            return [np .nan ,np .nan ,np .nan ,area ,width ,np .nan ,np .nan ,np .nan ,np .nan ]
        highest_peak =max (peaks ,key =lambda peak :signal [peak ])
        height =signal [highest_peak ]
        fwhm_ ,heightatfwhm_ ,_ ,_ =sig .peak_widths (signal ,[highest_peak ],rel_height =0.5 )
        fwhm =fwhm_ [0 ]*sample_period 
        heightatfwhm =heightatfwhm_ [0 ]
        skew =stats .skew (signal )
        kurt =stats .kurtosis (signal )

        event_time =event_signal [0 ]/sampling_rate 
        if np .abs (height )>np .abs (event_baseline_mean ):
            print ("Ignoring event as height is greater than baseline")
            return [np .nan ,np .nan ,np .nan ,np .nan ,np .nan ,np .nan ,np .nan ,np .nan ,np .nan ]
        return [height ,fwhm ,heightatfwhm ,area ,width ,skew ,kurt ,event_baseline_mean ,event_time ]

    else :
        if ML_standard =="Scheme 1":
            signal =np .abs (event_signal )

            sample_period =1 /sampling_rate 
            width =len (signal )*sample_period 
            n =10 
            part_averages =np .full (n ,np .nan )


            samples_per_part =len (signal )//n if n >0 else 0 


            num_parts =min (n ,len (signal )//samples_per_part )if samples_per_part >0 else 0 


            part_width =width /num_parts if num_parts >0 else width 


            for i in range (num_parts ):
                start_index =i *samples_per_part 
                end_index =(i +1 )*samples_per_part 

                part_samples =signal [start_index :end_index ]

                part_averages [i ]=np .mean (part_samples )


            return np .append (part_averages ,part_width )

        elif ML_standard =="Scheme 2":
            signal =np .abs (event_signal )


            sample_period =1 /sampling_rate 
            width =len (signal )*sample_period 
            n =10 
            part_averages =np .full (n ,np .nan )


            samples_per_part =len (signal )//n if n >0 else 0 


            num_parts =min (n ,len (signal )//samples_per_part )if samples_per_part >0 else 0 


            part_width =width /num_parts if num_parts >0 else width 


            for i in range (num_parts ):
                start_index =i *samples_per_part 
                end_index =(i +1 )*samples_per_part 

                part_samples =signal [start_index :end_index ]

                part_averages [i ]=np .mean (part_samples )

            area =np .sum (signal )*sample_period 
            peaks ,_ =sig .find_peaks (signal )
            if len (peaks )==0 :
                return np .full (16 ,np .nan )
            highest_peak =max (peaks ,key =lambda peak :signal [peak ])
            height =signal [highest_peak ]
            fwhm_ ,heightatfwhm_ ,_ ,_ =sig .peak_widths (signal ,[highest_peak ],rel_height =0.5 )
            fwhm =fwhm_ [0 ]*sample_period 
            heightatfwhm =heightatfwhm_ [0 ]
            return np .append (part_averages ,[part_width ,height ,fwhm ,heightatfwhm ,area ,width ])

        elif ML_standard =="Scheme 3":

            signal =np .abs (event_signal )

            sample_period =1 /sampling_rate 
            width =len (signal )*sample_period 
            n =10 
            part_averages =np .full (n ,np .nan )


            samples_per_part =len (signal )//n if n >0 else 0 


            num_parts =min (n ,len (signal )//samples_per_part )if samples_per_part >0 else 0 


            part_width =width /num_parts if num_parts >0 else width 


            for i in range (num_parts ):
                start_index =i *samples_per_part 
                end_index =(i +1 )*samples_per_part 

                part_samples =signal [start_index :end_index ]

                part_averages [i ]=np .mean (part_samples )

            area =np .sum (signal )*sample_period 
            peaks ,_ =sig .find_peaks (signal )
            if len (peaks )==0 :
                return np .full (18 ,np .nan )
            highest_peak =max (peaks ,key =lambda peak :signal [peak ])
            height =signal [highest_peak ]
            fwhm_ ,heightatfwhm_ ,_ ,_ =sig .peak_widths (signal ,[highest_peak ],rel_height =0.5 )
            fwhm =fwhm_ [0 ]*sample_period 
            heightatfwhm =heightatfwhm_ [0 ]
            skew =stats .skew (signal )
            kurt =stats .kurtosis (signal )
            return np .append (part_averages ,[part_width ,height ,fwhm ,heightatfwhm ,area ,width ,skew ,kurt ])

        elif ML_standard =="Scheme 4":
            signal =np .abs (event_signal )

            sample_period =1 /sampling_rate 
            width =len (signal )*sample_period 
            n =0 
            part_averages =np .full (n ,np .nan )


            samples_per_part =len (signal )//n if n >0 else 0 


            num_parts =min (n ,len (signal )//samples_per_part )if samples_per_part >0 else 0 


            part_width =width /num_parts if num_parts >0 else width 


            for i in range (num_parts ):
                start_index =i *samples_per_part 
                end_index =(i +1 )*samples_per_part 

                part_samples =signal [start_index :end_index ]

                part_averages [i ]=np .mean (part_samples )

            area =np .sum (signal )*sample_period 
            peaks ,_ =sig .find_peaks (signal )
            if len (peaks )==0 :
                return np .full (58 ,np .nan )
            highest_peak =max (peaks ,key =lambda peak :signal [peak ])
            height =signal [highest_peak ]
            fwhm_ ,heightatfwhm_ ,_ ,_ =sig .peak_widths (signal ,[highest_peak ],rel_height =0.5 )
            fwhm =fwhm_ [0 ]*sample_period 
            heightatfwhm =heightatfwhm_ [0 ]
            skew =stats .skew (signal )
            kurt =stats .kurtosis (signal )
            return np .append (part_averages ,[part_width ,height ,fwhm ,heightatfwhm ,area ,width ,skew ,kurt ])

        elif ML_standard =="Scheme 5":

            signal =np .abs (event_signal )

            sample_period =1 /sampling_rate 
            width =len (signal )*sample_period 
            area =np .sum (signal )*sample_period 
            peaks ,_ =sig .find_peaks (signal )

            max_points =int (sampling_rate /1000 )+8 
            result =np .full (max_points ,np .nan )

            if len (peaks )==0 :
                result [-8 :]=[np .nan ,np .nan ,np .nan ,area ,width ,np .nan ,np .nan ,event_baseline_mean ]
                return result 

            highest_peak =max (peaks ,key =lambda peak :signal [peak ])
            height =signal [highest_peak ]
            fwhm_ ,heightatfwhm_ ,_ ,_ =sig .peak_widths (signal ,[highest_peak ],rel_height =0.5 )
            fwhm =fwhm_ [0 ]*sample_period 
            heightatfwhm =heightatfwhm_ [0 ]
            skew =stats .skew (signal )
            kurt =stats .kurtosis (signal )

            signal_length =min (len (signal ),max_points -8 )
            result [:signal_length ]=signal [:signal_length ]
            result [-8 :]=[height ,fwhm ,heightatfwhm ,area ,width ,skew ,kurt ,event_baseline_mean ]

            return result 

def analyze_events_chunk (chunk ,data ,sampling_rate ,ML_standardisation_settings ):
    return [analyze_event (event ['event_data'],event ['baseline_value'],sampling_rate ,ML_standardisation_settings )for event in chunk ]

def save_chunked_event_analysis_to_npz (chunk_size ,data ,events ,sampling_rate ,ML_standardisation_settings ):
    event_chunks =np .array_split (events ,max (1 ,len (events )//chunk_size ))
    analysis_results =Parallel (n_jobs =-1 )(
    delayed (analyze_events_chunk )(chunk ,data ,sampling_rate ,ML_standardisation_settings )for chunk in event_chunks 
    )

    flattened_results =np .array ([item for sublist in analysis_results for item in sublist ],dtype =np .float64 )
    return flattened_results 

def process_chunk (data_chunk ,algorithm ,num_clusters_determination ,num_clusters_spinbox_value ,
max_clusters_silhouette_spinbox_value ,max_clusters_db_spinbox_value ,
threshold_spinbox_value ,covariance_type_combo_value ,linkage_combo_value ,
damping_spinbox_value ,bandwidth_spinbox_value ):


    data_chunk =np .array (data_chunk )


    data_chunk =data_chunk [~np .isnan (data_chunk ).any (axis =1 ),:]


    if data_chunk .shape [0 ]==0 :
        return np .array ([]),None 

    if algorithm =="K-Means Clustering":
        if num_clusters_determination =="Manual":
            num_clusters =min (num_clusters_spinbox_value ,data_chunk .shape [0 ])
        elif num_clusters_determination =="Silhouette Method":
            max_clusters =min (max_clusters_silhouette_spinbox_value ,data_chunk .shape [0 ])
            best_score =-1 
            best_num_clusters =2 
            for num_clusters in range (2 ,max_clusters +1 ):
                if num_clusters >=data_chunk .shape [0 ]:
                    break 
                kmeans =KMeans (n_clusters =num_clusters ,n_init ='auto')
                labels =kmeans .fit_predict (data_chunk )
                if len (np .unique (labels ))<2 :
                    continue 
                score =silhouette_score (data_chunk ,labels )
                if score >best_score :
                    best_score =score 
                    best_num_clusters =num_clusters 
            num_clusters =min (best_num_clusters ,data_chunk .shape [0 ]-1 )
        elif num_clusters_determination =="Davies-Bouldin Index":
            max_clusters =min (max_clusters_db_spinbox_value ,data_chunk .shape [0 ])
            best_score =float ('inf')
            best_num_clusters =2 
            for num_clusters in range (2 ,max_clusters +1 ):
                if num_clusters >=data_chunk .shape [0 ]:
                    break 
                kmeans =KMeans (n_clusters =num_clusters ,n_init ='auto')
                labels =kmeans .fit_predict (data_chunk )
                if len (np .unique (labels ))<2 :
                    continue 
                score =davies_bouldin_score (data_chunk ,labels )
                if score <best_score :
                    best_score =score 
                    best_num_clusters =num_clusters 
            num_clusters =min (best_num_clusters ,data_chunk .shape [0 ]-1 )
        else :
            threshold =threshold_spinbox_value 
            max_clusters =min (20 ,data_chunk .shape [0 ])
            num_clusters =2 


            max_distance_data =0 
            for i in range (data_chunk .shape [0 ]):
                for j in range (i +1 ,data_chunk .shape [0 ]):
                    distance =np .linalg .norm (data_chunk [i ]-data_chunk [j ])
                    if distance >max_distance_data :
                        max_distance_data =distance 

            while num_clusters <max_clusters :
                kmeans =KMeans (n_clusters =num_clusters ,n_init ='auto')
                labels =kmeans .fit_predict (data_chunk )
                centroids =kmeans .cluster_centers_ 

                max_distance_centroid =0 
                for i in range (data_chunk .shape [0 ]):
                    distance =np .linalg .norm (data_chunk [i ]-centroids [labels [i ]])
                    if distance >max_distance_centroid :
                        max_distance_centroid =distance 

                if max_distance_centroid <=threshold *max_distance_data :
                    break 

                num_clusters +=1 

            if num_clusters ==max_clusters :
                print ("Threshold not satisfied. Using maximum number of clusters.")

        if num_clusters >0 :
            kmeans =KMeans (n_clusters =num_clusters ,n_init ='auto')
            labels =kmeans .fit_predict (data_chunk )
            centroids =kmeans .cluster_centers_ 
        else :
            labels =np .array ([])
            centroids =None 

        return labels ,centroids 


    elif algorithm =="GMM":
        if num_clusters_determination =="Manual":
            num_components =num_clusters_spinbox_value 
            num_components =min (num_components ,data_chunk .shape [0 ])
        elif num_clusters_determination =="Silhouette Method":
            max_clusters =max_clusters_silhouette_spinbox_value 
            max_clusters =min (max_clusters ,data_chunk .shape [0 ])
            best_score =-1 
            best_num_components =2 
            for num_components in range (2 ,max_clusters +1 ):
                gmm =GaussianMixture (n_components =num_components ,covariance_type =covariance_type_combo_value )
                labels =gmm .fit_predict (data_chunk )
                if len (np .unique (labels ))<2 :
                    continue 
                score =silhouette_score (data_chunk ,labels )
                if score >best_score :
                    best_score =score 
                    best_num_components =num_components 
        elif num_clusters_determination =="Davies-Bouldin Index":
            max_clusters =max_clusters_db_spinbox_value 
            max_clusters =min (max_clusters ,data_chunk .shape [0 ])
            best_score =float ('inf')
            best_num_components =2 
            for num_components in range (2 ,max_clusters +1 ):
                gmm =GaussianMixture (n_components =num_components ,covariance_type =covariance_type_combo_value )
                labels =gmm .fit_predict (data_chunk )
                if len (np .unique (labels ))<2 :
                    continue 
                score =davies_bouldin_score (data_chunk ,labels )
                if score <best_score :
                    best_score =score 
                    best_num_components =num_components 
        else :
            threshold =threshold_spinbox_value 
            max_clusters =20 
            num_components =2 


            max_distance_data =0 
            for i in range (data_chunk .shape [0 ]):
                for j in range (i +1 ,data_chunk .shape [0 ]):
                    distance =np .linalg .norm (data_chunk [i ]-data_chunk [j ])
                    if distance >max_distance_data :
                        max_distance_data =distance 

            while num_components <max_clusters :
                gmm =GaussianMixture (n_components =num_components ,covariance_type =covariance_type_combo_value )
                labels =gmm .fit_predict (data_chunk )
                centroids =gmm .means_ 

                max_distance_centroid =0 
                for i in range (data_chunk .shape [0 ]):
                    distance =np .linalg .norm (data_chunk [i ]-centroids [labels [i ]])
                    if distance >max_distance_centroid :
                        max_distance_centroid =distance 

                if max_distance_centroid <=threshold *max_distance_data :
                    break 

                num_components +=1 

            if num_components ==max_clusters :
                print ("Threshold not satisfied. Using maximum number of components.")

        best_num_components =min (num_components ,data_chunk .shape [0 ])

        if best_num_components >0 :
            gmm =GaussianMixture (n_components =best_num_components ,covariance_type =covariance_type_combo_value )
            labels =gmm .fit_predict (data_chunk )
        else :
            labels =np .array ([])

        return labels ,None 

    elif algorithm =="Hierarchical Clustering":
        linkage =linkage_combo_value 
        if num_clusters_determination =="Manual":
            num_clusters =num_clusters_spinbox_value 
            num_clusters =min (num_clusters ,data_chunk .shape [0 ])
            hierarchical =AgglomerativeClustering (n_clusters =num_clusters ,linkage =linkage )
            labels =hierarchical .fit_predict (data_chunk )
            return labels ,None 
        else :
            if num_clusters_determination =="Silhouette Method":
                max_clusters =max_clusters_silhouette_spinbox_value 
                max_clusters =min (max_clusters ,data_chunk .shape [0 ])
                best_score =-1 
                best_num_clusters =2 
                for num_clusters in range (2 ,max_clusters +1 ):
                    if num_clusters >=data_chunk .shape [0 ]:
                        break 
                    hierarchical =AgglomerativeClustering (n_clusters =num_clusters ,linkage =linkage )
                    labels =hierarchical .fit_predict (data_chunk )
                    if len (np .unique (labels ))<2 :
                        continue 
                    score =silhouette_score (data_chunk ,labels )
                    if score >best_score :
                        best_score =score 
                        best_num_clusters =num_clusters 
                best_num_clusters =min (best_num_clusters ,data_chunk .shape [0 ]-1 )
            elif num_clusters_determination =="Davies-Bouldin Index":
                max_clusters =max_clusters_db_spinbox_value 
                max_clusters =min (max_clusters ,data_chunk .shape [0 ])
                best_score =float ('inf')
                best_num_clusters =2 
                for num_clusters in range (2 ,max_clusters +1 ):
                    if num_clusters >=data_chunk .shape [0 ]:
                        break 
                    hierarchical =AgglomerativeClustering (n_clusters =num_clusters ,linkage =linkage )
                    labels =hierarchical .fit_predict (data_chunk )
                    if len (np .unique (labels ))<2 :
                        continue 
                    score =davies_bouldin_score (data_chunk ,labels )
                    if score <best_score :
                        best_score =score 
                        best_num_clusters =num_clusters 
                best_num_clusters =min (best_num_clusters ,data_chunk .shape [0 ]-1 )
            else :
                threshold =threshold_spinbox_value 
                max_clusters =max_clusters_silhouette_spinbox_value 
                max_clusters =min (max_clusters ,data_chunk .shape [0 ])
                num_clusters =2 
                best_num_clusters =num_clusters 
                while num_clusters <max_clusters :
                    hierarchical =AgglomerativeClustering (n_clusters =num_clusters ,linkage =linkage )
                    labels =hierarchical .fit_predict (data_chunk )
                    if len (np .unique (labels ))<2 :
                        num_clusters +=1 
                        continue 
                    score =silhouette_score (data_chunk ,labels )
                    if score >=threshold :
                        best_num_clusters =num_clusters 
                        break 
                    num_clusters +=1 

        if best_num_clusters >0 :
            hierarchical =AgglomerativeClustering (n_clusters =best_num_clusters ,linkage =linkage )
            labels =hierarchical .fit_predict (data_chunk )
        else :
            labels =np .array ([])

        return labels ,None 

    elif algorithm =="Affinity Propagation":
        damping =damping_spinbox_value 
        affinity =AffinityPropagation (damping =damping ,max_iter =2000 )
        labels =affinity .fit_predict (data_chunk )

        return labels ,None 

    elif algorithm =="Mean Shift":
        bandwidth =bandwidth_spinbox_value 
        mean_shift =MeanShift (bandwidth =bandwidth )
        labels =mean_shift .fit_predict (data_chunk )

        return labels ,None 

class SDEventClusteringApp (QWidget ):
    def __init__ (self ):
        super ().__init__ ()
        self .setWindowTitle ("SD Event Clustering and Data Reduction App")
        self .setGeometry (100 ,100 ,1200 ,800 )
        self .data =None 
        self .events_data ={}

        main_layout =QHBoxLayout ()
        self .setLayout (main_layout )

        left_layout =QVBoxLayout ()
        right_layout =QVBoxLayout ()
        main_layout .addLayout (left_layout ,3 )
        main_layout .addLayout (right_layout ,7 )


        self .title_label =QLabel ('SD Event Clustering and Data Reduction App')
        self .title_label .setFont (QFont ('Arial',23 ,QFont .Weight .Bold ))
        self .title_label .setAlignment (Qt .AlignmentFlag .AlignCenter )
        self .subtitle_label =QLabel ('shankar.dutt@anu.edu.au')
        self .subtitle_label .setFont (QFont ('Arial',15 ,QFont .Weight .Bold ))
        self .subtitle_label .setAlignment (Qt .AlignmentFlag .AlignCenter )
        left_layout .addWidget (self .title_label )
        left_layout .addWidget (self .subtitle_label )

        self .tabs =QTabWidget ()
        left_layout .addWidget (self .tabs )


        load_files_tab =QWidget ()
        load_files_layout =QVBoxLayout (load_files_tab )

        self .select_folder_btn =QPushButton ("Select Folder")
        self .select_folder_btn .clicked .connect (self .select_folder )
        self .include_subfolders_chk =QCheckBox ("Include Subfolders")
        self .files_list_widget =QListWidget ()

        self .files_list_widget .setSelectionMode (QListWidget .SelectionMode .SingleSelection )
        self .folder_path_label =QLabel (" ")
        self .folder_path_label .setWordWrap (True )
        load_files_layout .addWidget (self .select_folder_btn )
        load_files_layout .addWidget (self .include_subfolders_chk )
        load_files_layout .addWidget (self .files_list_widget )
        load_files_layout .addWidget (self .folder_path_label )

        self .perform_clustering_btn =QPushButton ("Perform Clustering")
        self .perform_clustering_btn .clicked .connect (self .perform_clustering )
        self .perform_clustering_data_reduction_btn =QPushButton ("Perform Clustering and Data Reduction")
        self .perform_clustering_data_reduction_btn .clicked .connect (self .perform_clustering_data_reduction )
        load_files_layout .addWidget (self .perform_clustering_btn )
        load_files_layout .addWidget (self .perform_clustering_data_reduction_btn )

        self .tabs .addTab (load_files_tab ,"Load Files")


        clustering_settings_tab =QWidget ()
        clustering_settings_layout =QVBoxLayout (clustering_settings_tab )

        scroll_area =QScrollArea ()
        scroll_area .setWidgetResizable (True )
        scroll_content =QWidget ()
        scroll_layout =QVBoxLayout (scroll_content )
        scroll_area .setWidget (scroll_content )
        clustering_settings_layout .addWidget (scroll_area )


        self .settings_group =QGroupBox ("Clustering Settings")
        self .settings_layout =QFormLayout (self .settings_group )
        scroll_layout .addWidget (self .settings_group )


        self .event_type_label =QLabel ("Event Type:")
        self .event_type_combo =QComboBox ()
        self .event_type_combo .addItems (["Dips","Peaks"])
        self .event_type_combo .setCurrentIndex (0 )
        self .settings_layout .addRow (self .event_type_label ,self .event_type_combo )



        alignment_padding_group =QGroupBox ("Alignment and Padding Settings")
        alignment_padding_layout =QFormLayout (alignment_padding_group )


        self .alignment_method_label =QLabel ("Alignment Method:")
        self .alignment_method_combo =QComboBox ()
        self .alignment_method_combo .addItems (["Maximum","Minimum","Center"])
        self .alignment_method_combo .setCurrentIndex (1 )
        alignment_padding_layout .addRow (self .alignment_method_label ,self .alignment_method_combo )


        self .padding_method_label =QLabel ("Padding Method:")
        self .padding_method_combo =QComboBox ()
        self .padding_method_combo .addItems (["NaN","Zero"])
        self .padding_method_combo .setCurrentIndex (1 )
        alignment_padding_layout .addRow (self .padding_method_label ,self .padding_method_combo )

        scroll_layout .addWidget (alignment_padding_group )


        algorithm_group =QGroupBox ("Clustering Algorithm")
        algorithm_layout =QFormLayout (algorithm_group )
        self .algorithm_combo =QComboBox ()
        self .algorithm_combo .addItems (["K-Means Clustering","GMM","Hierarchical Clustering",
        "Affinity Propagation","Mean Shift"])
        self .algorithm_combo .currentIndexChanged .connect (self .update_algorithm_settings )
        algorithm_layout .addRow ("Algorithm:",self .algorithm_combo )
        scroll_layout .addWidget (algorithm_group )

        self .settings_group =QGroupBox ("Clustering Settings")
        self .settings_layout =QFormLayout (self .settings_group )



        self .kmeans_max_clusters_label =QLabel ("Max Clusters (K-Means):")
        self .kmeans_max_clusters_spinbox =QSpinBox ()
        self .kmeans_max_clusters_spinbox .setRange (2 ,1000 )
        self .kmeans_max_clusters_spinbox .setValue (20 )
        self .settings_layout .addRow (self .kmeans_max_clusters_label ,self .kmeans_max_clusters_spinbox )


        self .gmm_max_clusters_label =QLabel ("Max Clusters (GMM):")
        self .gmm_max_clusters_spinbox =QSpinBox ()
        self .gmm_max_clusters_spinbox .setRange (2 ,1000 )
        self .gmm_max_clusters_spinbox .setValue (20 )
        self .settings_layout .addRow (self .gmm_max_clusters_label ,self .gmm_max_clusters_spinbox )


        self .hierarchical_max_clusters_label =QLabel ("Max Clusters (Hierarchical):")
        self .hierarchical_max_clusters_spinbox =QSpinBox ()
        self .hierarchical_max_clusters_spinbox .setRange (2 ,1000 )
        self .hierarchical_max_clusters_spinbox .setValue (20 )
        self .settings_layout .addRow (self .hierarchical_max_clusters_label ,self .hierarchical_max_clusters_spinbox )
        self .hierarchical_max_clusters_label .setVisible (False )
        self .hierarchical_max_clusters_spinbox .setVisible (False )


        self .num_clusters_determination_label =QLabel ("Number of Clusters Determination:")
        self .num_clusters_determination_combo =QComboBox ()
        self .num_clusters_determination_combo .addItems (["Manual","Silhouette Method","Davies-Bouldin Index","Threshold"])
        self .num_clusters_determination_combo .currentIndexChanged .connect (self .update_num_clusters_determination )
        self .settings_layout .addRow (self .num_clusters_determination_label ,self .num_clusters_determination_combo )


        self .num_clusters_label =QLabel ("Number of Clusters:")
        self .num_clusters_spinbox =QSpinBox ()
        self .num_clusters_spinbox .setRange (1 ,1000 )
        self .num_clusters_spinbox .setValue (5 )
        self .num_clusters_label .setVisible (False )
        self .num_clusters_spinbox .setVisible (False )
        self .settings_layout .addRow (self .num_clusters_label ,self .num_clusters_spinbox )


        self .max_clusters_silhouette_label =QLabel ("Max Clusters (Silhouette):")
        self .max_clusters_silhouette_spinbox =QSpinBox ()
        self .max_clusters_silhouette_spinbox .setRange (2 ,1000 )
        self .max_clusters_silhouette_spinbox .setValue (20 )
        self .max_clusters_silhouette_label .setVisible (False )
        self .max_clusters_silhouette_spinbox .setVisible (False )
        self .settings_layout .addRow (self .max_clusters_silhouette_label ,self .max_clusters_silhouette_spinbox )


        self .max_clusters_db_label =QLabel ("Max Clusters (Davies-Bouldin):")
        self .max_clusters_db_spinbox =QSpinBox ()
        self .max_clusters_db_spinbox .setRange (2 ,1000 )
        self .max_clusters_db_spinbox .setValue (20 )
        self .max_clusters_db_label .setVisible (False )
        self .max_clusters_db_spinbox .setVisible (False )
        self .settings_layout .addRow (self .max_clusters_db_label ,self .max_clusters_db_spinbox )


        self .threshold_label =QLabel ("Threshold:")
        self .threshold_spinbox =QDoubleSpinBox ()
        self .threshold_spinbox .setRange (0.1 ,1.0 )
        self .threshold_spinbox .setSingleStep (0.05 )
        self .threshold_spinbox .setValue (0.85 )
        self .threshold_label .setVisible (False )
        self .threshold_spinbox .setVisible (False )
        self .settings_layout .addRow (self .threshold_label ,self .threshold_spinbox )


        self .covariance_type_label =QLabel ("Covariance Type:")
        self .covariance_type_combo =QComboBox ()
        self .covariance_type_combo .addItems (["full","tied","diag","spherical"])
        self .covariance_type_label .setVisible (False )
        self .covariance_type_combo .setVisible (False )
        self .settings_layout .addRow (self .covariance_type_label ,self .covariance_type_combo )


        self .damping_label =QLabel ("Damping:")
        self .damping_spinbox =QDoubleSpinBox ()
        self .damping_spinbox .setRange (0.5 ,1.0 )
        self .damping_spinbox .setSingleStep (0.05 )
        self .damping_spinbox .setValue (0.5 )
        self .damping_label .setVisible (False )
        self .damping_spinbox .setVisible (False )
        self .settings_layout .addRow (self .damping_label ,self .damping_spinbox )


        self .bandwidth_label =QLabel ("Bandwidth:")
        self .bandwidth_spinbox =QDoubleSpinBox ()
        self .bandwidth_spinbox .setRange (0.1 ,1000.0 )
        self .bandwidth_spinbox .setSingleStep (0.1 )
        self .bandwidth_spinbox .setValue (100.0 )
        self .bandwidth_label .setVisible (False )
        self .bandwidth_spinbox .setVisible (False )
        self .settings_layout .addRow (self .bandwidth_label ,self .bandwidth_spinbox )


        self .linkage_label =QLabel ("Linkage:")
        self .linkage_combo =QComboBox ()
        self .linkage_combo .addItems (["ward","complete","average","single"])
        self .linkage_label .setVisible (False )
        self .linkage_combo .setVisible (False )
        self .settings_layout .addRow (self .linkage_label ,self .linkage_combo )

        scroll_layout .addWidget (self .settings_group )

        self .tabs .addTab (clustering_settings_tab ,"Clustering Settings")


        data_reduction_settings_tab =QWidget ()
        data_reduction_settings_layout =QVBoxLayout (data_reduction_settings_tab )

        scroll_area =QScrollArea ()
        scroll_area .setWidgetResizable (True )
        scroll_content =QWidget ()
        scroll_layout =QVBoxLayout (scroll_content )
        scroll_area .setWidget (scroll_content )
        data_reduction_settings_layout .addWidget (scroll_area )

        normal_reduction_group =QGroupBox ("Standardisation")
        normal_reduction_layout =QFormLayout (normal_reduction_group )
        self .enable_standardisation_checkbox =QCheckBox ("Enable Standardisation")
        self .enable_standardisation_checkbox .stateChanged .connect (self .toggle_standardisation_widgets )
        normal_reduction_layout .addWidget (self .enable_standardisation_checkbox )
        self .standardisation_type_label =QLabel ("Which Standardisation to use:")
        self .standardisation_type_combo =QComboBox ()
        self .standardisation_type_combo .addItems (["ΔI/I₀","(ΔI*I₀)⁰·⁵","(ΔI*I₀)ᵖᵒʷᵉʳ","Dutt Standardisation"])
        self .standardisation_type_combo .currentIndexChanged .connect (self .update_standardisation_explanation )
        normal_reduction_layout .addRow (self .standardisation_type_label ,self .standardisation_type_combo )
        self .standardisation_explanation_label =QLabel ()
        self .standardisation_explanation_label .setWordWrap (True )
        normal_reduction_layout .addRow (self .standardisation_explanation_label )
        self .power_label =QLabel ("Power:")
        self .power_spinbox =QDoubleSpinBox ()
        self .power_spinbox .setRange (0.01 ,10 )
        self .power_spinbox .setValue (0.5 )
        self .power_spinbox .setSingleStep (0.01 )
        self .power_label .setVisible (False )
        self .power_spinbox .setVisible (False )
        normal_reduction_layout .addRow (self .power_label ,self .power_spinbox )
        self .length_label =QLabel ("Length of the nanopore (L) (nm):")
        self .length_spinbox =QDoubleSpinBox ()
        self .length_spinbox .setRange (1 ,100 )
        self .length_spinbox .setValue (7 )
        self .length_spinbox .setSingleStep (0.1 )
        self .length_label .setVisible (False )
        self .length_spinbox .setVisible (False )
        normal_reduction_layout .addRow (self .length_label ,self .length_spinbox )
        self .conductivity_label =QLabel ("σ (Conductivity of the solution) (S/m):")
        self .conductivity_spinbox =QDoubleSpinBox ()
        self .conductivity_spinbox .setRange (0.1 ,100 )
        self .conductivity_spinbox .setValue (10.5 )
        self .conductivity_spinbox .setSingleStep (0.1 )
        self .conductivity_label .setVisible (False )
        self .conductivity_spinbox .setVisible (False )
        normal_reduction_layout .addRow (self .conductivity_label ,self .conductivity_spinbox )
        self .voltage_label =QLabel ("V (Voltage Applied) (mV):")
        self .voltage_spinbox =QDoubleSpinBox ()
        self .voltage_spinbox .setRange (10 ,2000 )
        self .voltage_spinbox .setValue (400 )
        self .voltage_spinbox .setSingleStep (1 )
        self .voltage_label .setVisible (False )
        self .voltage_spinbox .setVisible (False )
        normal_reduction_layout .addRow (self .voltage_label ,self .voltage_spinbox )
        self .open_pore_current_label =QLabel ("I₀ (Open Pore Current) (nA):")
        self .open_pore_current_spinbox =QDoubleSpinBox ()
        self .open_pore_current_spinbox .setRange (-500 ,500 )
        self .open_pore_current_spinbox .setValue (25 )
        self .open_pore_current_spinbox .setSingleStep (0.1 )
        self .open_pore_current_label .setVisible (False )
        self .open_pore_current_spinbox .setVisible (False )
        normal_reduction_layout .addRow (self .open_pore_current_label ,self .open_pore_current_spinbox )
        scroll_layout .addWidget (normal_reduction_group )

        self .check_values_group =QGroupBox ("Check Values")
        self .check_values_layout =QFormLayout (self .check_values_group )
        self .delta_i_label =QLabel ("ΔI (Change in Current) (nA):")
        self .delta_i_spinbox =QDoubleSpinBox ()
        self .delta_i_spinbox .setRange (-500 ,500 )
        self .delta_i_spinbox .setValue (1 )
        self .delta_i_spinbox .setSingleStep (0.1 )
        self .check_values_layout .addRow (self .delta_i_label ,self .delta_i_spinbox )
        self .check_values_button =QPushButton ("Check Values")
        self .check_values_button .clicked .connect (self .check_values )
        self .check_values_layout .addRow (self .check_values_button )
        self .check_values_result_label =QLabel ()
        self .check_values_layout .addRow (self .check_values_result_label )
        self .check_values_group .setVisible (False )
        scroll_layout .addWidget (self .check_values_group )

        ml_reduction_group =QGroupBox ("ML Data Reduction")
        ml_reduction_layout =QFormLayout (ml_reduction_group )
        self .enable_ml_data_reduction_checkbox =QCheckBox ("Enable ML Data Reduction")
        self .enable_ml_data_reduction_checkbox .stateChanged .connect (self .toggle_ml_widgets )
        ml_reduction_layout .addWidget (self .enable_ml_data_reduction_checkbox )
        self .tag_label =QLabel ("Tag:")
        self .tag_lineedit =QLineEdit ()
        self .tag_lineedit .setText ("BSA")
        ml_reduction_layout .addRow (self .tag_label ,self .tag_lineedit )
        self .data_reduction_type_label =QLabel ("Type of Data Reduction:")
        self .data_reduction_type_combo =QComboBox ()
        self .data_reduction_type_combo .addItems (["Scheme 1","Scheme 2","Scheme 3","Scheme 4","Scheme 5"])
        ml_reduction_layout .addRow (self .data_reduction_type_label ,self .data_reduction_type_combo )
        self .scheme_table =QTableWidget ()
        self .scheme_table .setRowCount (5 )
        self .scheme_table .setColumnCount (2 )
        self .scheme_table .setHorizontalHeaderLabels (["Feature Extraction Scheme","Features/Details"])
        self .scheme_table .setItem (0 ,0 ,QTableWidgetItem ("Scheme 1"))
        self .scheme_table .setItem (0 ,1 ,QTableWidgetItem ("Δi1, Δi2, … Δi10, Δt0/10"))
        self .scheme_table .setItem (1 ,0 ,QTableWidgetItem ("Scheme 2"))
        self .scheme_table .setItem (1 ,1 ,QTableWidgetItem ("Δi1, Δi2, … Δi10, Δt0/10, Δtfwhm, area, Δifwhm, Δimax"))
        self .scheme_table .setItem (2 ,0 ,QTableWidgetItem ("Scheme 3"))
        self .scheme_table .setItem (2 ,1 ,QTableWidgetItem ("Δi1, Δi2, … Δi10, Δt0/10, Δtfwhm, area, Δifwhm, Δimax , skew, kurtosis"))
        self .scheme_table .setItem (3 ,0 ,QTableWidgetItem ("Scheme 4"))
        self .scheme_table .setItem (3 ,1 ,QTableWidgetItem ("Δi1, Δi2, … Δi50, Δt0/50, Δtfwhm, area, Δifwhm, Δimax , skew, kurtosis"))
        self .scheme_table .setItem (4 ,0 ,QTableWidgetItem ("Scheme 5"))
        self .scheme_table .setItem (4 ,1 ,QTableWidgetItem ("Full signal, Δtfwhm, area, Δifwhm, Δimax , skew, kurtosis"))
        self .scheme_table .resizeColumnsToContents ()
        ml_reduction_layout .addRow (self .scheme_table )
        scroll_layout .addWidget (ml_reduction_group )

        self .tabs .addTab (data_reduction_settings_tab ,"Data Reduction Settings")


        self .canvas =FigureCanvas (plt .Figure (figsize =(15 ,6 )))
        self .ax =self .canvas .figure .subplots ()
        right_layout .addWidget (self .canvas )

        toolbar =NavigationToolbar (self .canvas ,self )
        right_layout .addWidget (toolbar )

        self .result_text =QLabel ()
        self .result_text .setWordWrap (True )
        scroll_area =QScrollArea ()
        scroll_area .setWidgetResizable (True )
        scroll_area .setWidget (self .result_text )
        right_layout .addWidget (scroll_area )

        self .standardisation_widgets =[
        self .standardisation_type_label ,self .standardisation_type_combo ,
        self .standardisation_explanation_label ,self .power_label ,self .power_spinbox ,
        self .length_label ,self .length_spinbox ,self .conductivity_label ,self .conductivity_spinbox ,
        self .voltage_label ,self .voltage_spinbox ,self .open_pore_current_label ,
        self .open_pore_current_spinbox ,self .check_values_group 
        ]
        self .toggle_standardisation_widgets ()

        self .ml_widgets =[self .tag_label ,self .tag_lineedit ,self .data_reduction_type_label ,
        self .data_reduction_type_combo ,self .scheme_table ]
        self .toggle_ml_widgets ()

        self .update_algorithm_settings (0 )

    def toggle_standardisation_widgets (self ):
        enabled =self .enable_standardisation_checkbox .isChecked ()
        for widget in self .standardisation_widgets :
            widget .setVisible (enabled )
        if enabled :
            self .update_standardisation_explanation (self .standardisation_type_combo .currentIndex ())
        else :
            self .power_label .setVisible (False )
            self .power_spinbox .setVisible (False )
            self .length_label .setVisible (False )
            self .length_spinbox .setVisible (False )
            self .conductivity_label .setVisible (False )
            self .conductivity_spinbox .setVisible (False )
            self .voltage_label .setVisible (False )
            self .voltage_spinbox .setVisible (False )
            self .open_pore_current_label .setVisible (False )
            self .open_pore_current_spinbox .setVisible (False )
            self .check_values_group .setVisible (False )

    def update_standardisation_explanation (self ,index ):
        self .power_label .setVisible (index ==2 )
        self .power_spinbox .setVisible (index ==2 )
        self .length_label .setVisible (index ==3 )
        self .length_spinbox .setVisible (index ==3 )
        self .conductivity_label .setVisible (index ==3 )
        self .conductivity_spinbox .setVisible (index ==3 )
        self .voltage_label .setVisible (index ==3 )
        self .voltage_spinbox .setVisible (index ==3 )
        self .open_pore_current_label .setVisible (index ==3 )
        self .open_pore_current_spinbox .setVisible (index ==3 )
        self .check_values_group .setVisible (index ==3 )
        if index ==0 :
            self .standardisation_explanation_label .setText ("ΔI<sub>new</sub> = ΔI/I<sub>0</sub>")
        elif index ==1 :
            self .standardisation_explanation_label .setText ("ΔI<sub>new</sub> = (ΔI*I<sub>0</sub>)<sup>0.5</sup>")
        elif index ==2 :
            self .standardisation_explanation_label .setText ("ΔI<sub>new</sub> = (ΔI*I<sub>0</sub>)<sup>power</sup><br>"
            "Beware of using this option as some parts of ΔI<sub>new</sub> "
            "would be favoured as compared to others with changing values "
            "of power. Also beware of the units as except for power = 0.5, "
            "the units are not nA.")
        elif index ==3 :
            self .standardisation_explanation_label .setText ("For more information on this, talk with Shankar Dutt "
            "(shankar.dutt@anu.edu.au)")

    def check_values (self ):
        l =self .length_spinbox .value ()*10 **(-9 )
        delta_i =self .delta_i_spinbox .value ()*10 **(-9 )
        v =self .voltage_spinbox .value ()/1000 
        sigma =self .conductivity_spinbox .value ()
        i0 =self .open_pore_current_spinbox .value ()*10 **(-9 )
        d =i0 +(i0 *(i0 +16 *l *v *sigma /np .pi ))**0.5 /(2 *v *sigma )
        condition =sigma >(4 *l *delta_i /v +d *np .pi *delta_i /v )/(d **2 *np .pi )
        condition1 =delta_i <i0 
        if condition and condition1 :
            self .check_values_result_label .setText ("The entered values should work, and you will get a good standardisation.")
        else :
            self .check_values_result_label .setText ("The entered values will not work. Something is wrong!")
        self .check_values_result_label .setWordWrap (True )

    def toggle_ml_widgets (self ):
        for widget in self .ml_widgets :
            widget .setVisible (self .enable_ml_data_reduction_checkbox .isChecked ())

    def update_algorithm_settings (self ,index ):
        if index ==0 :
            self .num_clusters_label .setVisible (False )
            self .num_clusters_spinbox .setVisible (False )
            self .covariance_type_label .setVisible (False )
            self .covariance_type_combo .setVisible (False )
            self .damping_label .setVisible (False )
            self .damping_spinbox .setVisible (False )
            self .bandwidth_label .setVisible (False )
            self .bandwidth_spinbox .setVisible (False )
            self .linkage_label .setVisible (False )
            self .linkage_combo .setVisible (False )
            self .kmeans_max_clusters_label .setVisible (True )
            self .kmeans_max_clusters_spinbox .setVisible (True )
            self .gmm_max_clusters_label .setVisible (False )
            self .gmm_max_clusters_spinbox .setVisible (False )
            self .hierarchical_max_clusters_label .setVisible (False )
            self .hierarchical_max_clusters_spinbox .setVisible (False )
            self .num_clusters_label .setVisible (self .num_clusters_determination_combo .currentIndex ()==0 )
            self .num_clusters_spinbox .setVisible (self .num_clusters_determination_combo .currentIndex ()==0 )
            self .kmeans_max_clusters_label .setVisible (False )
            self .kmeans_max_clusters_spinbox .setVisible (False )
        elif index ==1 :
            self .num_clusters_label .setVisible (False )
            self .num_clusters_spinbox .setVisible (False )
            self .covariance_type_label .setVisible (True )
            self .covariance_type_combo .setVisible (True )
            self .damping_label .setVisible (False )
            self .damping_spinbox .setVisible (False )
            self .bandwidth_label .setVisible (False )
            self .bandwidth_spinbox .setVisible (False )
            self .linkage_label .setVisible (False )
            self .linkage_combo .setVisible (False )
            self .kmeans_max_clusters_label .setVisible (False )
            self .kmeans_max_clusters_spinbox .setVisible (False )
            self .gmm_max_clusters_label .setVisible (True )
            self .gmm_max_clusters_spinbox .setVisible (True )
            self .hierarchical_max_clusters_label .setVisible (False )
            self .hierarchical_max_clusters_spinbox .setVisible (False )
        elif index ==2 :
            self .num_clusters_label .setVisible (False )
            self .num_clusters_spinbox .setVisible (False )
            self .covariance_type_label .setVisible (False )
            self .covariance_type_combo .setVisible (False )
            self .damping_label .setVisible (False )
            self .damping_spinbox .setVisible (False )
            self .bandwidth_label .setVisible (False )
            self .bandwidth_spinbox .setVisible (False )
            self .linkage_label .setVisible (True )
            self .linkage_combo .setVisible (True )
            self .kmeans_max_clusters_label .setVisible (False )
            self .kmeans_max_clusters_spinbox .setVisible (False )
            self .gmm_max_clusters_label .setVisible (False )
            self .gmm_max_clusters_spinbox .setVisible (False )
            self .hierarchical_max_clusters_label .setVisible (False )
            self .hierarchical_max_clusters_spinbox .setVisible (False )
            self .num_clusters_determination_label .setVisible (True )
            self .num_clusters_determination_combo .setVisible (True )
            self .num_clusters_label .setVisible (self .num_clusters_determination_combo .currentIndex ()==0 )
            self .num_clusters_spinbox .setVisible (self .num_clusters_determination_combo .currentIndex ()==0 )
        elif index ==3 :
            self .num_clusters_label .setVisible (False )
            self .num_clusters_spinbox .setVisible (False )
            self .covariance_type_label .setVisible (False )
            self .covariance_type_combo .setVisible (False )
            self .damping_label .setVisible (True )
            self .damping_spinbox .setVisible (True )
            self .bandwidth_label .setVisible (False )
            self .bandwidth_spinbox .setVisible (False )
            self .linkage_label .setVisible (False )
            self .linkage_combo .setVisible (False )
            self .num_clusters_determination_label .setVisible (False )
            self .num_clusters_determination_combo .setVisible (False )
        elif index ==4 :
            self .num_clusters_label .setVisible (False )
            self .num_clusters_spinbox .setVisible (False )
            self .covariance_type_label .setVisible (False )
            self .covariance_type_combo .setVisible (False )
            self .damping_label .setVisible (False )
            self .damping_spinbox .setVisible (False )
            self .bandwidth_label .setVisible (True )
            self .bandwidth_spinbox .setVisible (True )
            self .linkage_label .setVisible (False )
            self .linkage_combo .setVisible (False )
            self .num_clusters_determination_label .setVisible (False )
            self .num_clusters_determination_combo .setVisible (False )

    def update_num_clusters_determination (self ,index ):
        self .num_clusters_label .setVisible (index ==0 )
        self .num_clusters_spinbox .setVisible (index ==0 )
        self .max_clusters_silhouette_label .setVisible (index ==1 )
        self .max_clusters_silhouette_spinbox .setVisible (index ==1 )
        self .max_clusters_db_label .setVisible (index ==2 )
        self .max_clusters_db_spinbox .setVisible (index ==2 )
        self .threshold_label .setVisible (index ==3 )
        self .threshold_spinbox .setVisible (index ==3 )


        self .kmeans_max_clusters_label .setVisible (False )
        self .kmeans_max_clusters_spinbox .setVisible (False )
        self .gmm_max_clusters_label .setVisible (False )
        self .gmm_max_clusters_spinbox .setVisible (False )
        self .hierarchical_max_clusters_label .setVisible (False )
        self .hierarchical_max_clusters_spinbox .setVisible (False )

    def select_folder (self ):
        options =QFileDialog .Option .ShowDirsOnly 
        directory =QFileDialog .getExistingDirectory (self ,"Select Folder","",options =options )
        if directory :
            self .folder_path_label .setText (f"Selected folder: {directory}")
            self .populate_file_list (directory ,self .include_subfolders_chk .isChecked ())

    def populate_file_list (self ,directory ,include_subfolders ):
        self .files_list_widget .clear ()
        for root ,dirs ,files in os .walk (directory ):
            for file in files :
                if file .endswith ('.event_data.npz'):
                    rel_path =os .path .relpath (os .path .join (root ,file ),start =directory )
                    item =QListWidgetItem (rel_path )
                    item .setData (Qt .ItemDataRole .UserRole ,os .path .join (root ,file ))
                    self .files_list_widget .addItem (item )
            if not include_subfolders :
                break 
    def align_and_pad_events (self ,events_data ):
        alignment_indices =[]
        event_lengths =[]

        for event_data in events_data .values ():
            event =event_data ['event_data']
            if self .alignment_method_combo .currentText ()=="Maximum":
                alignment_index =np .argmax (event )
            elif self .alignment_method_combo .currentText ()=="Minimum":
                alignment_index =np .argmin (event )
            else :
                alignment_index =len (event )//2 
            alignment_indices .append (alignment_index )
            event_lengths .append (len (event ))

        max_length =max (event_lengths )
        max_left_pad =max (alignment_indices )
        max_right_pad =max (max_length -np .array (alignment_indices ))
        total_length =max_left_pad +max_right_pad 

        aligned_data =[]
        for event_data ,alignment_index in zip (events_data .values (),alignment_indices ):
            event =event_data ['event_data']
            left_pad =max_left_pad -alignment_index 
            right_pad =total_length -len (event )-left_pad 

            if self .padding_method_combo .currentText ()=="NaN":
                padded_event =np .pad (event ,(left_pad ,right_pad ),mode ='constant',constant_values =np .nan )
            else :
                padded_event =np .pad (event ,(left_pad ,right_pad ),mode ='constant',constant_values =0 )

            aligned_data .append (padded_event )

        return np .array (aligned_data )

    def load_data (self ):
        self .events_data ={}
        events_data =self .data ['events']
        event_type =self .event_type_combo .currentText ()

        filtered_count =0 
        for event_counter ,event_data in enumerate (events_data ):
            start_time =event_data ['start_time']
            end_time =event_data ['end_time']
            event_signal =event_data ['event_data']
            event_baseline_mean =event_data ['baseline_value']

            if event_type =="Dips":
                extremum =np .min (event_signal )
                if np .abs (extremum )<=np .abs (event_baseline_mean ):
                    self .events_data [event_counter ]={
                    'event_id':event_data ['event_id'],
                    'start_time':start_time ,
                    'end_time':end_time ,
                    'event_data':event_signal ,
                    'baseline_value':event_baseline_mean 
                    }
                else :
                    filtered_count +=1 
            else :
                extremum =np .min (event_signal )
                if np .abs (extremum )<=np .abs (event_baseline_mean ):
                    self .events_data [event_counter ]={
                    'event_id':event_data ['event_id'],
                    'start_time':start_time ,
                    'end_time':end_time ,
                    'event_data':event_signal ,
                    'baseline_value':event_baseline_mean 
                    }
                else :
                    filtered_count +=1 

        self .sampling_rate =self .data ['sampling_rate'].item ()
        print (f"Total events loaded: {len(self.events_data)}")
        print (f"Events filtered out: {filtered_count}")

    def plot_clusters (self ,aligned_data ,labels ,centroids ,algorithm ):
        self .ax .clear ()
        unique_labels =np .unique (labels )
        colors =plt .colormaps ['tab20b'](np .linspace (0 ,1 ,len (unique_labels )))
        if len (aligned_data )<1000 :
            alpha_small =0.05 
        elif len (aligned_data )<5000 :
            alpha_small =0.025 
        else :
            alpha_small =0.01 

        for i ,label in enumerate (unique_labels ):
            cluster_mask =(labels ==label )
            cluster_data =aligned_data [cluster_mask ]
            for event in cluster_data :
                time =np .arange (len (event ))/self .sampling_rate *1000 
                self .ax .plot (time ,event ,alpha =alpha_small ,color =colors [i ])
            if centroids is not None and i <len (centroids ):
                centroid =centroids [i ]
                time =np .arange (len (centroid ))/self .sampling_rate *1000 
                self .ax .plot (time ,centroid ,alpha =0.9 ,color =colors [i ],linewidth =2 ,label =f"Cluster {label}")

        self .ax .set_xlabel ('Time (ms)')
        self .ax .set_ylabel ('Amplitude')
        self .ax .set_title (f'{algorithm} Clustering')
        self .ax .grid (True )
        self .ax .legend (loc ='best')
        self .canvas .draw ()

    def perform_clustering (self ):
        self .data ={}
        self .events_data ={}
        selected_files =[item .data (Qt .ItemDataRole .UserRole )for item in self .files_list_widget .selectedItems ()]
        if not selected_files :

            QMessageBox .warning (self ,"No File Selected","No file was selected.")
            return 

        file_path =selected_files [0 ]
        try :
            npz_data =np .load (file_path ,allow_pickle =True )
            self .data ={key :npz_data [key ]for key in npz_data }
        except Exception as e :
            QMessageBox .critical (self ,"Error",f"Failed to load data: {str(e)}")
            return 

        self .load_data ()

        if self .enable_standardisation_checkbox .isChecked ():

            standard =self .standardisation_type_combo .currentText ()
            standard_power =self .power_spinbox .value ()
            standard_length_nm =self .length_spinbox .value ()
            standard_conductivity_S_m =self .conductivity_spinbox .value ()
            standard_voltage_applied_mV =self .voltage_spinbox .value ()
            standard_open_pore_current_nA =self .open_pore_current_spinbox .value ()


            for event_data in self .events_data .values ():
                event_data ['event_data']=standardize_events (
                [event_data ['event_data']],
                standard ,
                event_data ['baseline_value'],
                standard_power ,
                standard_length_nm ,
                standard_conductivity_S_m ,
                standard_voltage_applied_mV ,
                standard_open_pore_current_nA 
                )[0 ]


        if not self .events_data :
            print ("No valid event data found.")
            return 


        max_length =max (len (event_data ['event_data'])for event_data in self .events_data .values ())

        aligned_data =self .align_and_pad_events (self .events_data )

        algorithm =self .algorithm_combo .currentText ()
        num_clusters_determination =self .num_clusters_determination_combo .currentText ()

        num_chunks =mp .cpu_count ()
        chunk_size =len (aligned_data )//num_chunks 
        data_chunks =[aligned_data [i :i +chunk_size ]for i in range (0 ,len (aligned_data ),chunk_size )]


        progress_bar =tqdm (total =len (data_chunks ),desc ="Clustering Progress",unit ="chunk")


        results =Parallel (n_jobs =num_chunks )(delayed (process_chunk )(
        chunk ,algorithm ,num_clusters_determination ,
        self .num_clusters_spinbox .value (),self .max_clusters_silhouette_spinbox .value (),
        self .max_clusters_db_spinbox .value (),np .abs (1 -self .threshold_spinbox .value ()),
        self .covariance_type_combo .currentText (),self .linkage_combo .currentText (),
        self .damping_spinbox .value (),
        self .bandwidth_spinbox .value ()
        )for chunk in data_chunks )


        progress_bar .update (len (data_chunks ))
        progress_bar .close ()


        labels_list =[result [0 ]for result in results ]
        centroids_list =[result [1 ]for result in results if result [1 ]is not None ]

        if labels_list :
            labels =np .concatenate (labels_list )
        else :
            labels =np .array ([])

        if centroids_list :
            centroids =np .vstack (centroids_list )
        else :
            centroids =None 


        aligned_data =np .array (aligned_data )

        if len (labels )!=len (aligned_data ):

            missing_labels =np .full (len (aligned_data )-len (labels ),-1 )
            labels =np .concatenate ((labels ,missing_labels ))


        nan_mask =np .isnan (aligned_data ).any (axis =1 )
        aligned_data_no_nan =aligned_data [~nan_mask ]
        labels_no_nan =labels [~nan_mask ]

        self .plot_clusters (aligned_data ,labels ,centroids ,algorithm )



        if algorithm =="K-Means Clustering":
            if centroids is not None :
                centroids =np .vstack (centroids )
            self .ax .set_title (f"K-Means Clustering")
            if len (np .unique (labels_no_nan ))<2 :
                self .result_text .setText (f"Clustering Algorithm: K-Means\n"
                f"All data points are assigned to a single cluster.")
            else :
                self .result_text .setText (f"Clustering Algorithm: K-Means\n"
                f"Silhouette Score: {silhouette_score(aligned_data_no_nan, labels_no_nan):.2f}\n"
                f"Davies-Bouldin Index: {davies_bouldin_score(aligned_data_no_nan, labels_no_nan):.2f}")
        elif algorithm =="GMM":
            self .ax .set_title (f"GMM Clustering (covariance_type={self.covariance_type_combo.currentText()})")
            if len (np .unique (labels_no_nan ))<2 :
                self .result_text .setText (f"Clustering Algorithm: GMM\n"
                f"Covariance Type: {self.covariance_type_combo.currentText()}\n"
                f"All data points are assigned to a single cluster.")
            else :
                self .result_text .setText (f"Clustering Algorithm: GMM\n"
                f"Covariance Type: {self.covariance_type_combo.currentText()}\n"
                f"Silhouette Score: {silhouette_score(aligned_data_no_nan, labels_no_nan):.2f}\n"
                f"Davies-Bouldin Index: {davies_bouldin_score(aligned_data_no_nan, labels_no_nan):.2f}")
        elif algorithm =="Hierarchical Clustering":
            self .ax .set_title (f"Hierarchical Clustering (linkage={self.linkage_combo.currentText()})")
            if len (np .unique (labels ))<2 :
                self .result_text .setText (f"Clustering Algorithm: Hierarchical\n"
                f"Linkage: {self.linkage_combo.currentText()}\n"
                f"All data points are assigned to a single cluster.")
            else :
                self .result_text .setText (f"Clustering Algorithm: Hierarchical\n"
                f"Linkage: {self.linkage_combo.currentText()}\n"
                f"Silhouette Score: {silhouette_score(aligned_data, labels):.2f}\n"
                f"Davies-Bouldin Index: {davies_bouldin_score(aligned_data, labels):.2f}")
        elif algorithm =="Affinity Propagation":
            self .ax .set_title (f"Affinity Propagation (damping={self.damping_spinbox.value()}, )")
            if len (np .unique (labels ))<2 :
                self .result_text .setText (f"Clustering Algorithm: Affinity Propagation\n"
                f"Damping: {self.damping_spinbox.value()}\n"

                f"All data points are assigned to a single cluster.")
            else :
                self .result_text .setText (f"Clustering Algorithm: Affinity Propagation\n"
                f"Damping: {self.damping_spinbox.value()}\n"

                f"Silhouette Score: {silhouette_score(aligned_data, labels):.2f}\n"
                f"Davies-Bouldin Index: {davies_bouldin_score(aligned_data, labels):.2f}")
        elif algorithm =="Mean Shift":
            self .ax .set_title (f"Mean Shift (bandwidth={self.bandwidth_spinbox.value()})")
            if len (np .unique (labels ))<2 :
                self .result_text .setText (f"Clustering Algorithm: Mean Shift\n"
                f"Bandwidth: {self.bandwidth_spinbox.value()}\n"
                f"All data points are assigned to a single cluster.")
            else :
                self .result_text .setText (f"Clustering Algorithm: Mean Shift\n"
                f"Bandwidth: {self.bandwidth_spinbox.value()}\n"
                f"Silhouette Score: {silhouette_score(aligned_data, labels):.2f}\n"
                f"Davies-Bouldin Index: {davies_bouldin_score(aligned_data, labels):.2f}")

        self .canvas .draw ()


        unique_labels =np .unique (labels )
        num_clusters =len (unique_labels )

        cluster_info =""
        for label in unique_labels :
            cluster_mask =(labels ==label )
            cluster_data =aligned_data [cluster_mask ]
            cluster_size =len (cluster_data )
            cluster_centroid =np .mean (cluster_data ,axis =0 )

            cluster_info +=f"Cluster {label}:\n"
            cluster_info +=f"  Size: {cluster_size}\n"
            cluster_info +=f"  Centroid: {cluster_centroid}\n\n"

        self .result_text .setText (self .result_text .text ()+"\n\nCluster Information:\n"+cluster_info )

    def perform_clustering_data_reduction (self ):
        self .data ={}
        self .events_data ={}
        selected_files =[item .data (Qt .ItemDataRole .UserRole )for item in self .files_list_widget .selectedItems ()]
        if not selected_files :

            print ("No file selected.")
            return 

        file_path =selected_files [0 ]
        try :
            npz_data =np .load (file_path ,allow_pickle =True )
            self .data ={key :npz_data [key ]for key in npz_data }
        except Exception as e :
            QMessageBox .critical (self ,"Error",f"Failed to load data: {str(e)}")
            return 

        self .load_data ()

        if self .enable_standardisation_checkbox .isChecked ():

            standard =self .standardisation_type_combo .currentText ()
            standard_power =self .power_spinbox .value ()
            standard_length_nm =self .length_spinbox .value ()
            standard_conductivity_S_m =self .conductivity_spinbox .value ()
            standard_voltage_applied_mV =self .voltage_spinbox .value ()
            standard_open_pore_current_nA =self .open_pore_current_spinbox .value ()

            for event_data in self .events_data .values ():
                event_data ['event_data']=standardize_events (
                [event_data ['event_data']],
                standard ,
                event_data ['baseline_value'],
                standard_power ,
                standard_length_nm ,
                standard_conductivity_S_m ,
                standard_voltage_applied_mV ,
                standard_open_pore_current_nA 
                )[0 ]


        if not self .events_data :
            print ("No valid event data found.")
            return 


        max_length =max (len (event_data ['event_data'])for event_data in self .events_data .values ())

        aligned_data =self .align_and_pad_events (self .events_data )

        algorithm =self .algorithm_combo .currentText ()
        num_clusters_determination =self .num_clusters_determination_combo .currentText ()

        num_chunks =mp .cpu_count ()
        chunk_size =len (aligned_data )//num_chunks 
        data_chunks =[aligned_data [i :i +chunk_size ]for i in range (0 ,len (aligned_data ),chunk_size )]


        progress_bar =tqdm (total =len (data_chunks ),desc ="Clustering Progress",unit ="chunk")


        results =Parallel (n_jobs =num_chunks )(delayed (process_chunk )(
        chunk ,algorithm ,num_clusters_determination ,
        self .num_clusters_spinbox .value (),self .max_clusters_silhouette_spinbox .value (),
        self .max_clusters_db_spinbox .value (),np .abs (1 -self .threshold_spinbox .value ()),
        self .covariance_type_combo .currentText (),self .linkage_combo .currentText (),
        self .damping_spinbox .value (),
        self .bandwidth_spinbox .value ()
        )for chunk in data_chunks )


        progress_bar .update (len (data_chunks ))
        progress_bar .close ()


        labels_list =[result [0 ]for result in results ]
        centroids_list =[result [1 ]for result in results if result [1 ]is not None ]

        if labels_list :
            labels =np .concatenate (labels_list )
        else :
            labels =np .array ([])

        if centroids_list :
            centroids =np .vstack (centroids_list )
        else :
            centroids =None 


        aligned_data =np .array (aligned_data )

        print (f"Length of aligned_data: {len(aligned_data)}")
        print (f"Length of labels: {len(labels)}")

        if len (labels )!=len (aligned_data ):

            missing_labels =np .full (len (aligned_data )-len (labels ),-1 )
            labels =np .concatenate ((labels ,missing_labels ))


        nan_mask =np .isnan (aligned_data ).any (axis =1 )
        aligned_data_no_nan =aligned_data [~nan_mask ]
        labels_no_nan =labels [~nan_mask ]

        self .plot_clusters (aligned_data ,labels ,centroids ,algorithm )

        if algorithm =="K-Means Clustering":
            if centroids is not None :
                centroids =np .vstack (centroids )
            self .ax .set_title (f"K-Means Clustering")
            if len (np .unique (labels_no_nan ))<2 :
                self .result_text .setText (f"Clustering Algorithm: K-Means\n"
                f"All data points are assigned to a single cluster.")
            else :
                self .result_text .setText (f"Clustering Algorithm: K-Means\n"
                f"Silhouette Score: {silhouette_score(aligned_data_no_nan, labels_no_nan):.2f}\n"
                f"Davies-Bouldin Index: {davies_bouldin_score(aligned_data_no_nan, labels_no_nan):.2f}")
        elif algorithm =="GMM":
            self .ax .set_title (f"GMM Clustering (covariance_type={self.covariance_type_combo.currentText()})")
            if len (np .unique (labels_no_nan ))<2 :
                self .result_text .setText (f"Clustering Algorithm: GMM\n"
                f"Covariance Type: {self.covariance_type_combo.currentText()}\n"
                f"All data points are assigned to a single cluster.")
            else :
                self .result_text .setText (f"Clustering Algorithm: GMM\n"
                f"Covariance Type: {self.covariance_type_combo.currentText()}\n"
                f"Silhouette Score: {silhouette_score(aligned_data_no_nan, labels_no_nan):.2f}\n"
                f"Davies-Bouldin Index: {davies_bouldin_score(aligned_data_no_nan, labels_no_nan):.2f}")
        elif algorithm =="Hierarchical Clustering":
            self .ax .set_title (f"Hierarchical Clustering (linkage={self.linkage_combo.currentText()})")
            if len (np .unique (labels ))<2 :
                self .result_text .setText (f"Clustering Algorithm: Hierarchical\n"
                f"Linkage: {self.linkage_combo.currentText()}\n"
                f"All data points are assigned to a single cluster.")
            else :
                self .result_text .setText (f"Clustering Algorithm: Hierarchical\n"
                f"Linkage: {self.linkage_combo.currentText()}\n"
                f"Silhouette Score: {silhouette_score(aligned_data, labels):.2f}\n"
                f"Davies-Bouldin Index: {davies_bouldin_score(aligned_data, labels):.2f}")
        elif algorithm =="Affinity Propagation":
            self .ax .set_title (f"Affinity Propagation (damping={self.damping_spinbox.value()}, )")
            if len (np .unique (labels ))<2 :
                self .result_text .setText (f"Clustering Algorithm: Affinity Propagation\n"
                f"Damping: {self.damping_spinbox.value()}\n"
                f"All data points are assigned to a single cluster.")
            else :
                self .result_text .setText (f"Clustering Algorithm: Affinity Propagation\n"
                f"Damping: {self.damping_spinbox.value()}\n"
                f"Silhouette Score: {silhouette_score(aligned_data, labels):.2f}\n"
                f"Davies-Bouldin Index: {davies_bouldin_score(aligned_data, labels):.2f}")
        elif algorithm =="Mean Shift":
            self .ax .set_title (f"Mean Shift (bandwidth={self.bandwidth_spinbox.value()})")
            if len (np .unique (labels ))<2 :
                self .result_text .setText (f"Clustering Algorithm: Mean Shift\n"
                f"Bandwidth: {self.bandwidth_spinbox.value()}\n"
                f"All data points are assigned to a single cluster.")
            else :
                self .result_text .setText (f"Clustering Algorithm: Mean Shift\n"
                f"Bandwidth: {self.bandwidth_spinbox.value()}\n"
                f"Silhouette Score: {silhouette_score(aligned_data, labels):.2f}\n"
                f"Davies-Bouldin Index: {davies_bouldin_score(aligned_data, labels):.2f}")

        self .canvas .draw ()


        unique_labels =np .unique (labels )
        num_clusters =len (unique_labels )

        cluster_info =""
        for label in unique_labels :
            cluster_mask =(labels ==label )
            cluster_data =aligned_data [cluster_mask ]
            cluster_size =len (cluster_data )
            cluster_centroid =np .mean (cluster_data ,axis =0 )

            cluster_info +=f"Cluster {label}:\n"
            cluster_info +=f"  Size: {cluster_size}\n"
            cluster_info +=f"  Centroid: {cluster_centroid}\n\n"

        self .result_text .setText (self .result_text .text ()+"\n\nCluster Information:\n"+cluster_info )


        standard =self .standardisation_type_combo .currentText ()
        ML_enabled =self .enable_ml_data_reduction_checkbox .isChecked ()
        ML_standard =self .data_reduction_type_combo .currentText ()
        standard_power =self .power_spinbox .value ()
        standard_length_nm =self .length_spinbox .value ()
        standard_conductivity_S_m =self .conductivity_spinbox .value ()
        standard_voltage_applied_mV =self .voltage_spinbox .value ()
        standard_open_pore_current_nA =self .open_pore_current_spinbox .value ()

        ML_standardisation_settings ={
        'standard':standard ,
        'ML_enabled':str (ML_enabled ),
        'ML_standard':ML_standard ,
        'standard_power':standard_power ,
        'standard_length_nm':standard_length_nm ,
        'standard_conductivity_S_m':standard_conductivity_S_m ,
        'standard_voltage_applied_mV':standard_voltage_applied_mV ,
        'standard_open_pore_current_nA':standard_open_pore_current_nA 
        }

        unique_labels =np .unique (labels )
        sampling_rate =self .sampling_rate 


        if selected_files :
            file_path =selected_files [0 ]
            file_directory =os .path .dirname (file_path )
            file_name =os .path .basename (file_path )
            file_name_without_extension ,_ =os .path .splitext (file_name )
            file_name_without_extension ,_ =os .path .splitext (file_name_without_extension )
        else :
            file_directory =""
            file_name =""

        for label in unique_labels :
            cluster_mask =(labels ==label )
            cluster_data =aligned_data [cluster_mask ]

            cluster_events =[self .events_data [i ]for i in range (len (labels ))if labels [i ]==label and i in self .events_data ]

            reduced_data =save_chunked_event_analysis_to_npz (
            30 ,
            cluster_data ,
            cluster_events ,
            self .sampling_rate ,
            ML_standardisation_settings 
            )


            date_time_str =datetime .now ().strftime ("%Y_%m_%d__%H_%M_%S")
            file_name =f"{file_name_without_extension}_cluster_{label}_{date_time_str}"
            if ML_enabled :
                file_name +=".MLdataset"
            else :
                file_name +=".dataset"

            save_file_path =os .path .join (file_directory ,file_name )
            np .savez_compressed (save_file_path ,settings =ML_standardisation_settings ,X =reduced_data )

        QMessageBox .information (self ,"Data Reduction","Data reduction and saving completed.")

if __name__ =="__main__":
    app =QApplication ([])
    app .setStyle (QStyleFactory .create ('Fusion'))


    palette =QPalette ()
    palette .setColor (QPalette .ColorRole .Window ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .WindowText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Base ,QColor (25 ,25 ,25 ))
    palette .setColor (QPalette .ColorRole .AlternateBase ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ToolTipBase ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .ToolTipText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Text ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Button ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ButtonText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .BrightText ,Qt .GlobalColor .red )
    palette .setColor (QPalette .ColorRole .Link ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .Highlight ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .HighlightedText ,Qt .GlobalColor .black )
    app .setPalette (palette )

    window =SDEventClusteringApp ()
    window .showMaximized ()
    app .exec ()




from __future__ import print_function 

import sys 

from macholib ._cmdline import main as _main 
from macholib .mach_o import CPU_TYPE_NAMES ,MH_CIGAM_64 ,MH_MAGIC_64 ,get_cpu_subtype 
from macholib .MachO import MachO 

ARCH_MAP ={
("<","64-bit"):"x86_64",
("<","32-bit"):"i386",
(">","64-bit"):"ppc64",
(">","32-bit"):"ppc",
}


def print_file (fp ,path ):
    print (path ,file =fp )
    m =MachO (path )
    for header in m .headers :
        seen =set ()

        if header .MH_MAGIC ==MH_MAGIC_64 or header .MH_MAGIC ==MH_CIGAM_64 :
            sz ="64-bit"
        else :
            sz ="32-bit"

        arch =CPU_TYPE_NAMES .get (header .header .cputype ,header .header .cputype )

        subarch =get_cpu_subtype (header .header .cputype ,header .header .cpusubtype )

        print (
        "    [%s endian=%r size=%r arch=%r subarch=%r]"
        %(header .__class__ .__name__ ,header .endian ,sz ,arch ,subarch ),
        file =fp ,
        )
        for _idx ,_name ,other in header .walkRelocatables ():
            if other not in seen :
                seen .add (other )
                print ("\t"+other ,file =fp )
    print ("",file =fp )


def main ():
    print (
    "WARNING: 'macho_dump' is deprecated, use 'python -mmacholib dump' " "instead"
    )
    _main (print_file )


if __name__ =="__main__":
    try :
        sys .exit (main ())
    except KeyboardInterrupt :
        pass 



import os 


def check (cmd ,mf ):
    m =mf .findNode ("pylsp")
    if m is None or m .filename is None :
        return None 

    includes =["pylsp.__main__","pylsp.python_lsp"]

    root_dir =m .packagepath [0 ]
    files =os .scandir (os .path .join (root_dir ,"plugins"))
    for file in files :
        if file .name .endswith (".py"):
            includes .append (".".join (["pylsp","plugins",file .name [:-3 ]]))

    return {"includes":includes }
